import os
import random
from pathlib import Path
from typing import Union

import numpy as np
import pytest
import xarray as xr

from clisops import CONFIG
from clisops.exceptions import InvalidParameterValue
from clisops.ops.subset import Subset, subset
from clisops.parameter import (
    area_parameter,
    level_interval,
    level_series,
    time_components,
    time_interval,
    time_parameter,
    time_series,
)
from clisops.utils.dataset_utils import determine_lon_lat_range
from clisops.utils.output_utils import _format_time  # noqa


def _load_ds(fpath: Union[str, Path]):
    if isinstance(fpath, (str, Path)):
        if str(fpath).endswith("*.nc"):
            return xr.open_mfdataset(fpath)
        else:
            return xr.open_dataset(fpath)
    return xr.open_mfdataset(fpath)


def test_subset_no_params(tmpdir, check_output_nc, nimbus):
    """Test subset without area param."""
    result = subset(
        ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    check_output_nc(result)


def test_subset_time(nimbus, tmpdir, check_output_nc):
    """Tests clisops subset function with a time subset."""
    result = subset(
        ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
        time=time_interval("2005-01-01T00:00:00", "2020-12-30T00:00:00"),
        area=(0, -90.0, 360.0, 90.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    check_output_nc(result)


def test_subset_args_as_parameter_classes(nimbus, tmpdir, check_output_nc):
    """Tests clisops subset function with a time subset with the arguments as parameter classes from roocs-utils."""

    time = time_parameter.TimeParameter(
        time_interval("2000-01-01T00:00:00", "2020-12-30T00:00:00")
    )
    area = area_parameter.AreaParameter((0, -90.0, 360.0, 90.0))

    result = subset(
        ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
        time=time,
        area=area,
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    check_output_nc(result)


@pytest.mark.parametrize(
    "dset",
    [
        "ATLAS_v1_CMIP5",
        "ATLAS_v1_EOBS",
        "ATLAS_v1_ERA5",
        "ATLAS_v0_CORDEX_NAM",
        "ATLAS_v0_CMIP6",
    ],
)
def test_subset_ATLAS_datasets(tmpdir, dset, check_output_nc, mini_esgf_data):
    """Test temporal and spatial subset for several ATLAS datasets."""
    time = time_parameter.TimeParameter(
        time_interval("2000-01-01T00:00:00", "2020-12-30T00:00:00")
    )
    area = area_parameter.AreaParameter((0, -90.0, 360.0, 90.0))

    result = subset(
        ds=mini_esgf_data[dset],
        time=time,
        area=area,
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    check_output_nc(result)


def test_subset_invalid_time(nimbus, tmpdir):
    """Tests subset with invalid time param."""
    with pytest.raises(InvalidParameterValue):
        subset(
            ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
            time=time_interval("yesterday", "2020-12-30T00:00:00"),
            area=(0, -90.0, 360.0, 90.0),
            output_dir=tmpdir,
            output_type="nc",
            file_namer="simple",
        )


def test_subset_ds_is_none(tmpdir):
    """Tests subset with ds=None."""
    with pytest.raises(InvalidParameterValue):
        subset(
            ds=None,  # noqa
            time=time_interval("2020-01-01T00:00:00", "2020-12-30T00:00:00"),
            area=(0, -90.0, 360.0, 90.0),
            output_dir=tmpdir,
        )


def test_subset_no_ds(tmpdir):
    """Tests subset with no dataset provided."""
    with pytest.raises(TypeError):
        subset(
            time=time_interval("2020-01-01T00:00:00", "2020-12-30T00:00:00"),
            area=(0, -90.0, 360.0, 90.0),
            output_dir=tmpdir,
        )  # noqa


def test_subset_area_simple_file_name(nimbus, tmpdir, check_output_nc):
    """Tests clisops subset function with an area subset (simple file name)."""
    result = subset(
        ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
        area=(0.0, 10.0, 10.0, 65.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    check_output_nc(result)


def test_subset_area_project_file_name_atlas(tmpdir, check_output_nc, mini_esgf_data):
    """Tests clisops subset function with an area subset (derived file name)."""
    result = subset(
        ds=mini_esgf_data["ATLAS_v1_EOBS_GRID"],
        area=(0.0, 10.0, 10.0, 65.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="standard",
    )
    check_output_nc(result, "t_E-OBS_no-expt_mon_19500101-19500101.nc")


def test_subset_area_project_file_name(nimbus, tmpdir, check_output_nc):
    """Tests clisops subset function with an area subset (derived file name)."""
    result = subset(
        ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
        area=(0.0, 10.0, 10.0, 65.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="standard",
    )
    check_output_nc(result, "tas_mon_HadGEM2-ES_rcp85_r1i1p1_20051216-20301116.nc")


def test_subset_invalid_area(nimbus, tmpdir):
    """Tests subset with invalid area param."""
    with pytest.raises(InvalidParameterValue):
        subset(
            ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
            area=("zero", 49.0, 10.0, 65.0),
            output_dir=tmpdir,
        )


def test_subset_with_time_and_area(nimbus, tmpdir):
    """Tests clisops subset function with time and area subsets.

    On completion:
    - assert all dimensions have been reduced.
    """
    start_time, end_time = "2019-01-16", "2020-12-16"
    bbox = (0.0, -80, 170.0, 65.0)

    outputs = subset(
        ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
        time=time_interval(start_time, end_time),
        area=bbox,
        output_dir=tmpdir,
        output_type="xarray",
    )

    ds = outputs[0]

    assert _format_time(ds.time.values.min()) == start_time
    assert _format_time(ds.time.values.max()) == end_time

    assert ds.lon.values.tolist() == [0]
    assert ds.lat.values.tolist() == [35]


def test_subset_4D_data_all_argument_permutations(tmpdir, mini_esgf_data):
    """Tests clisops subset function with:
    - no args (collection only)
    - time only
    - level only
    - bbox only
    - time + level
    - time + bbox
    - level + bbox
    - time + level + bbox

    On completion:
    - Check the shape of the response

    """
    # Found in file:
    # times = ("2015-01-16 12", "MANY MORE", "2024-12-16 12") [120]
    # plevs = (100000, 92500, 85000, 70000, 60000, 50000, 40000, 30000, 25000,
    #          20000, 15000, 10000, 7000, 5000, 3000, 2000, 1000, 500, 100) [19]
    # lats = (-88.9277353522959, -25.9141861518467, 37.1202943109788) [3]
    # lons = (0, 63.28125, 126.5625, 189.84375, 253.125, 316.40625) [6]

    # Requested subset
    time_input = time_interval("2022-01-01", "2022-06-01")
    level_input = level_interval(1000, 1000)
    bbox_input = (0.0, -80, 170.0, 65.0)

    # Define a set of inputs and the resulting shape expected
    test_inputs = [
        ["coll only", (None, None, None)],
        ["time only", (time_input, None, None)],
        ["level only", (None, level_input, None)],
        ["bbox only", (None, None, bbox_input)],
        ["time & level", (time_input, level_input, None)],
        ["time & bbox", (time_input, None, bbox_input)],
        ["level & bbox", (None, level_input, bbox_input)],
        ["time, level & bbox", (time_input, level_input, bbox_input)],
    ]

    # Full data shape
    initial_shape = [120, 19, 3, 6]

    # Test each set of inputs, check the output shape (slice) is correct
    for _, inputs in test_inputs:
        expected_shape = initial_shape[:]
        tm, level, bbox = inputs

        if tm:
            expected_shape[0] = 5
        if level:
            expected_shape[1] = 1
        if bbox:
            expected_shape[2:4] = 2, 3

        outputs = subset(
            ds=mini_esgf_data["CMIP6_TA"],
            time=tm,
            area=bbox,
            level=level,
            output_dir=tmpdir,
            output_type="xarray",
        )

        ds = outputs[0]
        assert ds.ta.shape == tuple(expected_shape)


def test_subset_with_multiple_files_tas(tmpdir, check_output_nc, mini_esgf_data):
    """Tests with multiple tas files"""
    result = subset(
        ds=mini_esgf_data["CMIP5_TAS"],
        time=time_interval("2001-01-01T00:00:00", "2020-12-30T00:00:00"),
        area=(0.0, 0.0, 10.0, 65.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    check_output_nc(result)


def test_subset_with_multiple_files_zostoga(tmpdir, check_output_nc, mini_esgf_data):
    """Tests with multiple zostoga files"""
    result = subset(
        ds=mini_esgf_data["CMIP5_ZOSTOGA"],
        time=time_interval("2000-01-01T00:00:00", "2020-12-30T00:00:00"),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    check_output_nc(result)


def test_subset_with_multiple_files_rh(tmpdir, check_output_nc, mini_esgf_data):
    """Tests with multiple rh files"""
    result = subset(
        ds=mini_esgf_data["CMIP5_RH"],
        time=time_interval("2005-01-01T00:00:00", "2020-12-30T00:00:00"),
        area=(0, -90.0, 360.0, 90.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    check_output_nc(result)


def test_subset_with_tas_series(tmpdir, tas_series, check_output_nc):
    """Test with tas_series fixture"""
    result = subset(
        ds=tas_series(["20", "22", "25"]),
        time=time_interval("2000-07-01T00:00:00", "2020-12-30T00:00:00"),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    check_output_nc(result)


def test_time_slices_in_subset_tas(mini_esgf_data):
    start_time, end_time = "2001-01-01T00:00:00", "2200-12-30T00:00:00"

    time_slices = [
        ("2005-12-16", "2040-03-16"),
        ("2040-04-16", "2074-07-16"),
        ("2074-08-16", "2108-10-16"),
        ("2108-11-16", "2143-02-16"),
        ("2143-03-16", "2177-06-16"),
        ("2177-07-16", "2199-12-16"),
    ]

    config_max_file_size = CONFIG["clisops:write"]["file_size_limit"]
    temp_max_file_size = "10KB"
    CONFIG["clisops:write"]["file_size_limit"] = temp_max_file_size

    outputs = subset(
        ds=mini_esgf_data["CMIP5_TAS"],
        time=time_interval(start_time, end_time),
        area=(0.0, 5.0, 50.0, 90.0),
        output_type="xarray",
        file_namer="simple",
    )
    CONFIG["clisops:write"]["file_size_limit"] = config_max_file_size

    assert _format_time(outputs[0].time.values.min()) >= start_time
    assert _format_time(outputs[-1].time.values.max()) <= end_time

    count = 0
    for _ in outputs:
        assert _format_time(outputs[count].time.values.min()) >= time_slices[count][0]
        assert _format_time(outputs[count].time.values.max()) >= time_slices[count][1]
        count += 1


def test_time_slices_in_subset_rh(mini_esgf_data):
    start_time, end_time = "2001-01-01T00:00:00", "2200-12-30T00:00:00"

    time_slices = [
        ("2001-01-16", "2002-09-16"),
        ("2002-10-16", "2004-06-16"),
        ("2004-07-16", "2005-11-16"),
    ]

    config_max_file_size = CONFIG["clisops:write"]["file_size_limit"]
    temp_max_file_size = "10KB"
    CONFIG["clisops:write"]["file_size_limit"] = temp_max_file_size

    with xr.open_mfdataset(mini_esgf_data["CMIP5_RH"]) as ds:
        outputs = subset(
            ds=ds,
            time=time_interval(start_time, end_time),
            area=(0.0, 5.0, 50.0, 90.0),
            output_type="xarray",
            file_namer="simple",
        )

        CONFIG["clisops:write"]["file_size_limit"] = config_max_file_size

        assert _format_time(outputs[0].time.values.min()) >= start_time
        assert _format_time(outputs[-1].time.values.max()) <= end_time

        count = 0
        for _ in outputs:
            assert (
                _format_time(outputs[count].time.values.min()) >= time_slices[count][0]
            )
            assert (
                _format_time(outputs[count].time.values.max()) >= time_slices[count][1]
            )
            count += 1


# area can be a few degrees out
def test_area_within_area_subset(mini_esgf_data):
    area = (0.0, 10.0, 175.0, 90.0)

    outputs = subset(
        ds=mini_esgf_data["CMIP5_TAS"],
        time=time_interval("2001-01-01T00:00:00", "2200-12-30T00:00:00"),
        area=area,
        output_type="xarray",
        file_namer="simple",
    )

    ds = outputs[0]
    assert area[0] <= ds.lon.data <= area[2]
    assert area[1] <= ds.lat.data <= area[3]


def test_area_within_area_subset_cmip6(mini_esgf_data):
    area = (20.0, 10.0, 250.0, 90.0)

    outputs = subset(
        ds=mini_esgf_data["CMIP6_RLDS"],
        time=time_interval("2001-01-01T00:00:00", "2002-12-30T00:00:00"),
        area=area,
        output_type="xarray",
    )

    ds = outputs[0]

    assert area[0] <= ds.lon.data <= area[2]
    assert area[1] <= ds.lat.data <= area[3]
    assert ds.lon.data[0] == 250
    assert np.isclose(ds.lat.data[0], 36.76056)


def test_subset_with_lat_lon_single_values(mini_esgf_data):
    """Creates subset where lat and lon only have one value. Then
    subsets that. This tests that the `lat_bnds` and `lon_bnds`
    are not being reversed by the `_check_desc_coords` function in
    `clisops.core.subset`.
    """
    area = (20.0, 10.0, 250.0, 90.0)

    outputs = subset(
        ds=mini_esgf_data["CMIP6_RLDS"],
        time=time_interval("2001-01-01T00:00:00", "2002-12-30T00:00:00"),
        area=area,
        output_type="xarray",
    )

    ds = outputs[0]

    outputs2 = subset(
        ds=ds,
        time=time_interval("2001-01-01T00:00:00", "2002-12-30T00:00:00"),
        area=area,
        output_type="xarray",
    )

    ds2 = outputs2[0]
    assert len(ds2.lat) == 1
    assert len(ds2.lon) == 1


def test_area_within_area_subset_chunked(mini_esgf_data):
    start_time, end_time = "2001-01-01T00:00:00", "2200-12-30T00:00:00"
    area = (0.0, 10.0, 175.0, 90.0)

    config_max_file_size = CONFIG["clisops:write"]["file_size_limit"]
    temp_max_file_size = "10KB"
    CONFIG["clisops:write"]["file_size_limit"] = temp_max_file_size
    outputs = subset(
        ds=mini_esgf_data["CMIP5_TAS"],
        time=time_interval(start_time, end_time),
        area=area,
        output_type="xarray",
        file_namer="simple",
    )
    CONFIG["clisops:write"]["file_size_limit"] = config_max_file_size

    for ds in outputs:
        assert area[0] <= ds.lon.data <= area[2]
        assert area[1] <= ds.lat.data <= area[3]


def test_subset_level(nimbus):
    """Tests clisops subset function with a level subset."""
    # Levels are: 100000, ..., 100
    cmip6_o3 = nimbus.fetch(
        "cmip6/o3_Amon_GFDL-ESM4_historical_r1i1p1f1_gr1_185001-194912.nc",
    )

    ds = _load_ds(cmip6_o3)

    result1 = subset(
        ds=cmip6_o3, level=level_interval("100000/100"), output_type="xarray"
    )

    np.testing.assert_array_equal(result1[0].o3.values, ds.o3.values)

    result2 = subset(ds=cmip6_o3, level=level_interval("100/100"), output_type="xarray")

    np.testing.assert_array_equal(result2[0].o3.shape, (1200, 1, 2, 3))

    result3 = subset(
        ds=cmip6_o3, level=level_interval("101/-23.234"), output_type="xarray"
    )

    np.testing.assert_array_equal(result3[0].o3.values, result2[0].o3.values)


def test_aux_variables():
    """
    test auxiliary variables are remembered in output dataset
    Have to create a netcdf file with auxiliary variable
    """
    ds = _load_ds("tests/data/test_file.nc")

    assert "do_i_get_written" in ds.variables

    result = subset(
        ds=ds,
        time=time_interval("2005-01-01T00:00:00", "2020-12-30T00:00:00"),
        area=(0.0, 10.0, 10.0, 65.0),
        output_type="xarray",
    )

    assert "do_i_get_written" in result[0].variables


def test_coord_variables_exist(mini_esgf_data):
    """
    check coord variables e.g. lat/lon when original data
    is on an irregular grid exist in output dataset
    """
    ds = _load_ds(mini_esgf_data["C3S_CMIP5_TSICE"])

    assert "lat" in ds.coords
    assert "lon" in ds.coords

    result = subset(
        ds=mini_esgf_data["C3S_CMIP5_TSICE"],
        time=time_interval("2005-01-01T00:00:00", "2020-12-30T00:00:00"),
        area=(0.0, 10.0, 10.0, 65.0),
        output_type="xarray",
    )

    assert "lat" in result[0].coords
    assert "lon" in result[0].coords


def test_coord_variables_subsetted_i_j(mini_esgf_data):
    """
    check coord variables e.g. lat/lon when original data
    is on an irregular grid are subsetted correctly in output dataset
    """
    ds = _load_ds(mini_esgf_data["C3S_CMIP5_TSICE"])

    assert "lat" in ds.coords
    assert "lon" in ds.coords
    assert "i" in ds.dims
    assert "j" in ds.dims

    area = (50, -65.0, 250.0, 65.0)

    result = subset(
        ds=mini_esgf_data["C3S_CMIP5_TSICE"],
        time=time_interval("2005-01-01T00:00:00", "2020-12-30T00:00:00"),
        area=area,
        output_type="xarray",
    )

    out = result[0].tsice
    assert out.values.shape == (180, 7, 4)

    # all lats and lons (hence i and j) have been dropped in these ranges as they are all masked, only time dim remains
    assert out.where(
        np.logical_and(out.lon < area[0], out.lon > area[2]), drop=True
    ).values.shape == (180, 0, 0)
    assert out.where(
        np.logical_and(out.lat < area[1], out.lat > area[3]), drop=True
    ).values.shape == (180, 0, 0)

    mask1 = ~(np.isnan(out.sel(time=out.time[0])))

    assert np.all(out.lon.values[mask1.values] >= area[0])
    assert np.all(out.lon.values[mask1.values] <= area[2])
    assert np.all(out.lat.values[mask1.values] >= area[1])
    assert np.all(out.lat.values[mask1.values] <= area[3])


def test_coord_variables_subsetted_rlat_rlon(mini_esgf_data):
    """
    check coord variables e.g. lat/lon when original data
    is on an irregular grid are subsetted correctly in output dataset
    """
    ds = _load_ds(mini_esgf_data["CMIP5_WRONG_CF_UNITS"])

    assert "lat" in ds.coords
    assert "lon" in ds.coords
    assert "rlat" in ds.dims
    assert "rlon" in ds.dims

    area = (5.0, 10.0, 20.0, 65.0)

    result = subset(
        ds=mini_esgf_data["CMIP5_WRONG_CF_UNITS"],
        time=time_interval("2005-01-01T00:00:00", "2020-12-30T00:00:00"),
        area=area,
        output_type="xarray",
    )

    out = result[0].zos
    assert out.values.shape == (1, 65, 15)

    # all lats and lons (hence rlat and rlon) have been dropped in these ranges as they are all masked, only time dim remains
    assert out.where(
        np.logical_and(out.lon < area[0], out.lon > area[2]), drop=True
    ).values.shape == (1, 0, 0)
    assert out.where(
        np.logical_and(out.lat < area[1], out.lat > area[3]), drop=True
    ).values.shape == (1, 0, 0)

    mask1 = ~(np.isnan(out.sel(time=out.time[0])))

    assert np.all(out.lon.values[mask1.values] >= area[0])
    assert np.all(out.lon.values[mask1.values] <= area[2])
    assert np.all(out.lat.values[mask1.values] >= area[1])
    assert np.all(out.lat.values[mask1.values] <= area[3])


def test_time_invariant_subset_standard_name(tmpdir, check_output_nc, mini_esgf_data):
    result = subset(
        ds=mini_esgf_data["CMIP6_MRSOFC"],
        area=(5.0, 10.0, 360.0, 90.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="standard",
    )

    check_output_nc(result, fname="mrsofc_fx_IPSL-CM6A-LR_ssp119_r1i1p1f1_gr.nc")


def test_longitude_and_latitude_coords_only(tmpdir, check_output_nc, mini_esgf_data):
    """Test subset succeeds when latitude and longitude are coordinates not dims and are not called lat/lon"""
    result = subset(
        ds=mini_esgf_data["CMIP6_TOS"],
        area=(10, -70, 260, 70),
        output_dir=tmpdir,
        output_type="nc",
    )

    check_output_nc(
        result,
        fname="tos_Omon_MPI-ESM1-2-LR_historical_r1i1p1f1_gn_18500116-18691216.nc",
    )


def test_time_invariant_subset_simple_name(tmpdir, check_output_nc, mini_esgf_data):
    result = subset(
        ds=mini_esgf_data["CMIP6_MRSOFC"],
        area=(5.0, 10.0, 360.0, 90.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )

    check_output_nc(result)


def test_time_invariant_subset_with_time(mini_esgf_data):
    with pytest.raises(AttributeError) as exc:
        subset(
            ds=mini_esgf_data["CMIP6_MRSOFC"],
            time=time_interval("2005-01-01T00:00:00", "2020-12-30T00:00:00"),
            area=(5.0, 10.0, 360.0, 90.0),
            output_type="xarray",
        )
    assert str(exc.value) == "'Dataset' object has no attribute 'time'"


def test_cross_prime_meridian(tmpdir, mini_esgf_data, check_output_nc):
    """Test subset with crossing prime meridian"""
    ds = _load_ds(mini_esgf_data["CMIP6_TAS_DAY"])

    result = subset(
        ds=ds,
        area=(-5, 50, 30, 65),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )

    check_output_nc(result)


def test_do_not_cross_prime_meridian(tmpdir, mini_esgf_data, check_output_nc):
    """Test subset without crossing prime meridian"""
    ds = _load_ds(mini_esgf_data["CMIP6_TAS_DAY"])

    result = subset(
        ds=ds,
        area=(10, 50, 30, 65),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )

    check_output_nc(result)


def test_0_360_no_cross(tmpdir, mini_esgf_data, check_output_nc):
    ds = _load_ds(mini_esgf_data["CMIP6_RLDS_ONE_TIME_STEP"])

    result = subset(
        ds=ds,
        area=(10.0, -90.0, 200.0, 90.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )

    check_output_nc(result)


def test_0_360_cross(tmpdir, mini_esgf_data, check_output_nc):
    ds = _load_ds(mini_esgf_data["CMIP6_RLDS"])

    result = subset(
        ds=ds,
        area=(-50.0, -90.0, 100.0, 90.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )
    check_output_nc(result)


def test_300_60_no_cross(tmpdir, mini_esgf_data, check_output_nc):
    # longitude is -300 to 60
    ds = _load_ds(mini_esgf_data["CMIP6_AREACELLO"])

    result = subset(
        ds=ds,
        area=(10.0, -90.0, 50.0, 90.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )

    check_output_nc(result)


def test_300_60_cross(tmpdir, mini_esgf_data, check_output_nc):
    # longitude is -300 to 60
    ds = _load_ds(mini_esgf_data["CMIP6_AREACELLO"])

    result = subset(
        ds=ds,
        area=(-100.0, -90.0, 50.0, 90.0),
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )

    check_output_nc(result)


def test_roll_positive_real_data(mini_esgf_data):
    ds = _load_ds(mini_esgf_data["CMIP6_RLUS_ONE_TIME_STEP"])

    area = (-50.0, -90.0, 100.0, 90.0)

    result = subset(
        ds=ds,
        area=area,
        output_type="xarray",
    )

    assert result[0].lon.attrs == ds.lon.attrs

    assert area[0] <= all(result[0].lon.values) <= area[2]
    assert area[1] <= all(result[0].lat.values) <= area[3]

    # check array contains expected values
    assert np.array_equal(result[0].lon.values, np.arange(-50, 102.5, 2.5))


def test_roll_positive_mini_data(mini_esgf_data):
    ds = _load_ds(mini_esgf_data["CMIP6_TA"])

    area = (-180.0, -90.0, 120.0, 90.0)

    result = subset(
        ds=ds,
        area=area,
        output_type="xarray",
    )

    assert result[0].lon.attrs == ds.lon.attrs

    assert area[0] <= all(result[0].lon.values) <= area[2]
    assert area[1] <= all(result[0].lat.values) <= area[3]

    # check array contains expected values
    assert np.array_equal(
        result[0].lon.values, [-170.15625, -106.875, -43.59375, 0.0, 63.28125]
    )


def test_lon_alignment_curvilinear_grid(mini_esgf_data):
    ds = _load_ds(mini_esgf_data["CMIP6_SFTOF"])

    # Assert lon frame is (0, 360)
    x0, x1, *y = determine_lon_lat_range(
        ds,
        "longitude",
        "latitude",
        "vertices_longitude",
        "vertices_latitude",
        apply_fix=False,
    )
    assert np.isclose(x0, 0.0, atol=0.1)
    assert np.isclose(x1, 360.0, atol=0.1)

    # Define an area with another lon frame (-180, 180)
    area = (-50.0, -90.0, 100.0, 90.0)

    ds_subset = subset(
        ds=ds,
        area=area,
        output_type="xarray",
    )[0]

    # Assert data includes the desired area
    #  (lon_frame was converted to -180, 180 before subsetting)
    # Since it is a curvilinear grid, the minimum longitude is smaller than -50
    #  and the maximum longitude is larger than 100
    x0, x1, *y = determine_lon_lat_range(
        ds_subset,
        "longitude",
        "latitude",
        "longitude_bnds",
        "latitude_bnds",
        apply_fix=False,
    )
    assert np.isclose(x0, -155.0, atol=0.1)
    assert np.isclose(x1, 159.5, atol=0.1)

    # But we can confirm that the majority of points lie between -50 and 100
    #  (in this case more than 2/3 of the points)
    hist = np.histogram(ds_subset["longitude"], bins=(x0, -51, 101, x1))
    assert hist[0][1] > 2 * (hist[0][0] + hist[0][2])


class TestSubset:
    def test_resolve_params(self, nimbus):
        s = Subset(
            ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
            time=time_interval("1999-01-01T00:00:00", "2100-12-30T00:00:00"),
            area=(-5.0, 49.0, 10.0, 65),
            level=level_interval(1000.0, 1000.0),
        )

        assert s.params["start_date"] == "1999-01-01T00:00:00"
        assert s.params["end_date"] == "2100-12-30T00:00:00"
        assert s.params["lon_bnds"] == (-5, 10)
        assert s.params["lat_bnds"] == (49, 65)

    def test_resolve_params_time(self, nimbus):
        s = Subset(
            ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
            time=time_interval("1999-01-01", "2100-12"),
            area=(0, -90, 360, 90),
        )
        assert s.params["start_date"] == "1999-01-01T00:00:00"
        assert s.params["end_date"] == "2100-12-31T23:59:59"

    def test_resolve_params_invalid_time(self, nimbus):
        with pytest.raises(InvalidParameterValue):
            Subset(
                ds=nimbus.fetch(
                    "cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"
                ),
                time=time_interval("1999-01-01T00:00:00", "maybe tomorrow"),
                area=(0, -90, 360, 90),
            )
        with pytest.raises(InvalidParameterValue):
            Subset(
                ds=nimbus.fetch(
                    "cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"
                ),
                time=time_interval("", "2100"),
                area=(0, -90, 360, 90),
            )

    def test_resolve_params_area(self, nimbus):
        s = Subset(
            ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
            area=(0, 10, 50, 60),
        )
        assert s.params["lon_bnds"] == (0, 50)
        assert s.params["lat_bnds"] == (10, 60)
        # allow also strings
        s = Subset(
            ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
            area=("0", "10", "50", "60"),
        )
        assert s.params["lon_bnds"] == (0, 50)
        assert s.params["lat_bnds"] == (10, 60)

    def test_map_params_invalid_area(self, nimbus):
        with pytest.raises(InvalidParameterValue):
            Subset(
                ds=nimbus.fetch(
                    "cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"
                ),
                area=(0, 10, 50),
            )
        with pytest.raises(InvalidParameterValue):
            Subset(
                ds=nimbus.fetch(
                    "cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"
                ),
                area=("zero", 10, 50, 60),
            )


def test_end_date_nudged_backwards(mini_esgf_data):
    # use no leap dataset
    ds = _load_ds(mini_esgf_data["CMIP6_SICONC_DAY"])

    end_date = "2012-02-29T12:00:00"

    # check end date normally raises an error
    with pytest.raises(ValueError) as exc:
        ds.time.sel(time=slice(None, end_date))
    assert str(exc.value).startswith(
        "invalid day number provided in cftime.DatetimeNoLeap(2012, 2, 29, 12, 0, 0, 0"
    )

    result = subset(
        ds=mini_esgf_data["CMIP6_SICONC_DAY"],
        area=(20, 30.0, 150, 70.0),
        time=time_interval("2000-01-01T12:00:00", end_date),
        output_type="xarray",
    )

    # check end date of result is correct
    assert result[0].time.values[-1].strftime() == "2012-02-28 12:00:00"


def test_start_date_nudged_forwards(mini_esgf_data):
    # use no leap dataset
    ds = _load_ds(mini_esgf_data["CMIP6_SICONC_DAY"])

    start_date = "2012-02-29T12:00:00"

    # check start date normally raises an error
    with pytest.raises(ValueError) as exc:
        ds.time.sel(time=slice(None, start_date))
    assert str(exc.value).startswith(
        "invalid day number provided in cftime.DatetimeNoLeap(2012, 2, 29, 12, 0, 0, 0"
    )

    result = subset(
        ds=mini_esgf_data["CMIP6_SICONC_DAY"],
        area=(20, 30.0, 150, 70.0),
        time=time_interval(start_date, "2014-07-29T12:00:00"),
        output_type="xarray",
    )

    # check start date of result is correct
    assert result[0].time.values[0].strftime() == "2012-03-01 12:00:00"


def test_end_date_nudged_backwards_monthly_data(mini_esgf_data):
    # use no leap dataset
    ds = _load_ds(mini_esgf_data["CMIP6_SICONC"])

    end_date = "2012-02-29T12:00:00"

    # check end date normally raises an error
    with pytest.raises(ValueError) as exc:
        ds.time.sel(time=slice(None, end_date))
    assert str(exc.value).startswith(
        "invalid day number provided in cftime.DatetimeNoLeap(2012, 2, 29, 12, 0, 0, 0"
    )

    result = subset(
        ds=mini_esgf_data["CMIP6_SICONC"],
        area=(20, 30.0, 150, 70.0),
        time=time_interval("2000-01-01T12:00:00", end_date),
        output_type="xarray",
    )

    # check end date of result is correct
    assert result[0].time.values[-1].strftime() == "2012-02-15 00:00:00"


def test_start_date_nudged_backwards_monthly_data(mini_esgf_data):
    # use no leap dataset
    ds = _load_ds(mini_esgf_data["CMIP6_SICONC"])

    start_date = "2012-02-29T12:00:00"

    # check start date normally raises an error
    with pytest.raises(ValueError) as exc:
        ds.time.sel(time=slice(None, start_date))
    assert str(exc.value).startswith(
        "invalid day number provided in cftime.DatetimeNoLeap(2012, 2, 29, 12, 0, 0, 0"
    )

    result = subset(
        ds=mini_esgf_data["CMIP6_SICONC"],
        area=(20, 30.0, 150, 70.0),
        time=time_interval(start_date, "2014-07-29T12:00:00"),
        output_type="xarray",
    )

    # check start date of result is correct
    assert result[0].time.values[0].strftime() == "2012-03-16 12:00:00"


def test_no_lon_in_range(mini_esgf_data):
    with pytest.raises(Exception) as exc:
        subset(
            ds=mini_esgf_data["CMIP6_RLDS_ONE_TIME_STEP"],
            area=(8.37, -90, 8.56, 90),
            time=time_interval("2006-01-01T00:00:00", "2099-12-30T00:00:00"),
            output_type="xarray",
        )

    assert (
        str(exc.value)
        == "There were no valid data points found in the requested subset. Please expand "
        "the area covered by the bounding box, the time period or the level range you have selected."
    )


def test_no_lat_in_range(mini_esgf_data):
    with pytest.raises(Exception) as exc:
        subset(
            ds=mini_esgf_data["CMIP6_RLDS_ONE_TIME_STEP"],
            area=(0, 39.12, 360, 39.26),
            time=time_interval("2006-01-01T00:00:00", "2099-12-30T00:00:00"),
            output_type="xarray",
        )

    assert (
        str(exc.value)
        == "There were no valid data points found in the requested subset. Please expand "
        "the area covered by the bounding box, the time period or the level range you have selected."
    )


def test_no_lat_lon_in_range(mini_esgf_data):
    with pytest.raises(Exception) as exc:
        subset(
            ds=mini_esgf_data["CMIP6_RLDS_ONE_TIME_STEP"],
            area=(8.37, 39.12, 8.56, 39.26),
            time=time_interval("2006-01-01T00:00:00", "2099-12-30T00:00:00"),
            output_type="xarray",
        )

    assert (
        str(exc.value)
        == "There were no valid data points found in the requested subset. Please expand "
        "the area covered by the bounding box, the time period or the level range you have selected."
    )


def test_curvilinear_ds_no_data_in_bbox_real_data(mini_esgf_data):
    ds = _load_ds(mini_esgf_data["CMIP6_TOS_CNRM"])
    with pytest.raises(ValueError) as exc:
        subset(
            ds=ds,
            area="1,40,2,4",
            time=time_interval("2021-01-01/2050-12-31"),
            output_type="xarray",
        )
    assert (
        str(exc.value)
        == "There were no valid data points found in the requested subset. Please expand the area covered by the bounding box."
    )


def test_curvilinear_ds_no_data_in_bbox_real_data_swap_lat(mini_esgf_data):
    ds = _load_ds(mini_esgf_data["CMIP6_TOS_CNRM"])
    with pytest.raises(ValueError) as exc:
        subset(
            ds=ds,
            area="1,4,2,40",
            time=time_interval("2015-01-01/2050-12-31"),
            output_type="xarray",
        )
    assert (
        str(exc.value)
        == "There were no valid data points found in the requested subset. Please expand the area covered by the bounding box."
    )


def test_curvilinear_ds_no_data_in_bbox(mini_esgf_data):
    with pytest.raises(ValueError) as exc:
        subset(
            ds=mini_esgf_data["CMIP6_TOS_ONE_TIME_STEP"],
            area="1,5,1.2,4",
            time=time_interval("2021-01-01/2050-12-31"),
            output_type="xarray",
        )
    assert (
        str(exc.value)
        == "There were no valid data points found in the requested subset. "
        "Please expand the area covered by the bounding box."
    )


def test_curvilinear_increase_lon_of_bbox(mini_esgf_data):
    result = subset(
        ds=mini_esgf_data["CMIP6_TOS_ONE_TIME_STEP"],
        area="1,40,4,4",
        time=time_interval("2021-01-01/2050-12-31"),
        output_type="xarray",
    )

    assert result


class TestReverseBounds:

    rlds = "CMIP6_RLDS_ONE_TIME_STEP"
    tos = "CMIP6_TOS_ONE_TIME_STEP"

    def test_reverse_lat_regular(self, mini_esgf_data):
        result = subset(
            ds=mini_esgf_data[self.rlds],
            area=(20, -45, 240, 45),
            output_type="xarray",
        )

        result_rev = subset(
            ds=mini_esgf_data[self.rlds],
            area=(20, 45, 240, -45),
            output_type="xarray",
        )

        np.testing.assert_array_equal(result[0].rlds, result_rev[0].rlds)

    def test_reverse_lon_regular(self, mini_esgf_data):
        result = subset(
            ds=mini_esgf_data[self.rlds],
            area=(20, -45, 240, 45),
            output_type="xarray",
        )

        result_rev = subset(
            ds=mini_esgf_data[self.rlds],
            area=(240, -45, 20, 45),
            output_type="xarray",
        )

        np.testing.assert_array_equal(result[0].rlds, result_rev[0].rlds)

    def test_reverse_lon_cross_meridian_regular(self, mini_esgf_data):
        result = subset(
            ds=mini_esgf_data[self.rlds],
            area=(-70, -45, 240, 45),
            output_type="xarray",
        )

        result_rev = subset(
            ds=mini_esgf_data[self.rlds],
            area=(240, -45, -70, 45),
            output_type="xarray",
        )

        np.testing.assert_array_equal(result[0].rlds, result_rev[0].rlds)

    def test_reverse_lat_and_lon_regular(self, mini_esgf_data):
        result = subset(
            ds=mini_esgf_data[self.rlds],
            area=(-70, -45, 240, 45),
            output_type="xarray",
        )

        result_rev = subset(
            ds=mini_esgf_data[self.rlds],
            area=(240, 45, -70, -45),
            output_type="xarray",
        )

        np.testing.assert_array_equal(result[0].rlds, result_rev[0].rlds)

    def test_reverse_lat_curvilinear(self, mini_esgf_data):
        result = subset(
            ds=mini_esgf_data[self.tos],
            area=(20, -45, 240, 45),
            output_type="xarray",
        )

        result_rev = subset(
            ds=mini_esgf_data[self.tos],
            area=(20, 45, 240, -45),
            output_type="xarray",
        )

        np.testing.assert_array_equal(result[0].tos, result_rev[0].tos)

    def test_reverse_lon_curvilinear(self, mini_esgf_data):
        result = subset(
            ds=mini_esgf_data[self.tos],
            area=(20, -45, 240, 45),
            output_type="xarray",
        )

        result_rev = subset(
            ds=mini_esgf_data[self.tos],
            area=(240, -45, 20, 45),
            output_type="xarray",
        )

        np.testing.assert_array_equal(result[0].tos, result_rev[0].tos)

    def test_reverse_lon_cross_meridian_curvilinear(self, mini_esgf_data):
        result = subset(
            ds=mini_esgf_data[self.tos],
            area=(-70, -45, 120, 45),
            output_type="xarray",
        )

        result_rev = subset(
            ds=mini_esgf_data[self.tos],
            area=(120, -45, -70, 45),
            output_type="xarray",
        )

        np.testing.assert_array_equal(result[0].tos, result_rev[0].tos)

    def test_reverse_lat_and_lon_curvilinear(self, mini_esgf_data):
        result = subset(
            ds=mini_esgf_data[self.tos],
            area=(20, -45, 240, 45),
            output_type="xarray",
        )

        result_rev = subset(
            ds=mini_esgf_data[self.tos],
            area=(20, 45, 240, -45),
            output_type="xarray",
        )

        np.testing.assert_array_equal(result[0].tos, result_rev[0].tos)

    def test_reverse_with_desc_lat_lon_regular(self, mini_esgf_data):
        ds = _load_ds(mini_esgf_data[self.rlds])

        result = subset(
            ds=ds,
            area=(20, -45, 240, 45),
            output_type="xarray",
        )

        # make lat and lon descending
        ds_rev = ds.sortby("lat", ascending=False).sortby("lon", ascending=False)

        result_rev = subset(
            ds=ds_rev,
            area=(20, -45, 240, 45),
            output_type="xarray",
        )

        # return lat and lon to ascending
        result_rev = (
            result_rev[0].sortby("lat", ascending=True).sortby("lon", ascending=True)
        )

        np.testing.assert_array_equal(result[0].rlds, result_rev.rlds)

    def test_reverse_with_desc_lat_lon_curvilinear(self, mini_esgf_data):
        ds = _load_ds(mini_esgf_data[self.tos])

        result = subset(
            ds=ds,
            area=(20, -45, 240, 45),
            output_type="xarray",
        )

        # make i and j descending
        ds_rev = ds.sortby("i", ascending=False).sortby("j", ascending=False)

        result_rev = subset(
            ds=ds_rev,
            area=(20, -45, 240, 45),
            output_type="xarray",
        )

        # return lat and lon to ascending
        result_rev = (
            result_rev[0].sortby("i", ascending=True).sortby("j", ascending=True)
        )

        np.testing.assert_array_equal(result[0].tos, result_rev.tos)

    def test_reverse_level(self, nimbus):
        cmip6_o3 = nimbus.fetch(
            "cmip6/o3_Amon_GFDL-ESM4_historical_r1i1p1f1_gr1_185001-194912.nc",
        )

        result = subset(
            ds=cmip6_o3,
            level=level_interval("100000/100"),
            output_type="xarray",
        )

        result_rev = subset(
            ds=cmip6_o3,
            level=level_interval("100/100000"),
            output_type="xarray",
        )

        np.testing.assert_array_equal(result[0].o3, result_rev[0].o3)

    def test_reverse_time(self, mini_esgf_data):
        result = subset(
            ds=mini_esgf_data["CMIP5_TAS"],
            time=time_interval("2021-01-01/2050-12-31"),
            output_type="xarray",
        )

        assert result[0].time.size == 360

        with pytest.raises(ValueError) as exc:
            subset(
                ds=mini_esgf_data["CMIP5_TAS"],
                time=time_interval("2050-12-31/2021-01-01"),
                output_type="xarray",
            )
        assert (
            str(exc.value)
            == 'Start date ("2051-01-16T00:00:00") is after end date ("2020-12-16T00:00:00").'
        )


class TestByValues:

    @staticmethod
    def shuffle(lst):
        l_copy = lst[:]
        random.shuffle(l_copy)
        return l_copy

    @staticmethod
    def assert_vars_equal(var_id, *ds_list, extras=None):
        """Extract variable/DataArray `var_id` from each Dataset in the `ds_list`.

        Check they are all the same by comparing the arrays and common attributes.
        `extras` is an optional list of extra attributes to check.
        """
        if not extras:
            extras = []

        if len(ds_list) == 1:
            raise Exception("Only one Dataset passed to: _ds_var_check()")

        das = [ds[var_id] for ds in ds_list]
        ref_da = das[0]

        # Create a list of attributes to compare
        attrs = ["standard_name", "long_name", "units", "cell_methods"] + extras

        for da in das[1:]:
            assert da.values.tolist() == ref_da.values.tolist()

            for attr in attrs:
                if attr in ref_da.attrs:
                    assert ref_da.attrs[attr] == da.attrs.get(attr)

    def test_subset_level_by_values_all(self, tmpdir, mini_esgf_data):
        all_levels = [
            100000,
            92500,
            85000,
            70000,
            60000,
            50000,
            40000,
            30000,
            25000,
            20000,
            15000,
            10000,
            7000,
            5000,
            3000,
            2000,
            1000,
            500,
            100,
        ]

        shuffled_1 = self.shuffle(all_levels)
        shuffled_2 = self.shuffle(all_levels)

        # Get various outputs and compare they are the same
        ds_list = [
            subset(
                ds=mini_esgf_data["CMIP6_TA"],
                output_dir=tmpdir,
                output_type="xarray",
                level=level,
            )[0]
            for level in [
                None,
                level_series(all_levels),
                level_interval(all_levels[0], all_levels[-1]),
                level_series(list(reversed(all_levels))),
                level_series(shuffled_1),
                level_series(shuffled_2),
            ]
        ]

        self.assert_vars_equal("plev", *ds_list)

    def test_subset_level_by_values_partial(self, tmpdir, mini_esgf_data):
        some_levels = [
            60000,
            50000,
            40000,
            30000,
            25000,
            20000,
            15000,
            10000,
            7000,
            5000,
        ]

        shuffled_1 = self.shuffle(some_levels)
        shuffled_2 = self.shuffle(some_levels)

        # Get various outputs and compare they are the same
        ds_list = [
            subset(
                ds=mini_esgf_data["CMIP6_TA"],
                output_dir=tmpdir,
                output_type="xarray",
                level=level,
            )[0]
            for level in [
                level_series(some_levels),
                level_interval(some_levels[0], some_levels[-1]),
                level_series(list(reversed(some_levels))),
                level_series(shuffled_1),
                level_series(shuffled_2),
            ]
        ]

        self.assert_vars_equal("plev", *ds_list)

    def test_subset_level_by_values_with_gaps(self, tmpdir, mini_esgf_data):
        picked_levels = [60000, 30000, 25000, 20000, 7000, 5000]

        shuffled_1 = self.shuffle(picked_levels)
        shuffled_2 = self.shuffle(picked_levels)

        # Get various outputs and compare they are the same
        ds_list = [
            subset(
                ds=mini_esgf_data["CMIP6_TA"],
                output_dir=tmpdir,
                output_type="xarray",
                level=level,
            )[0]
            for level in [
                level_series(picked_levels),
                level_series(list(reversed(picked_levels))),
                level_series(shuffled_1),
                level_series(shuffled_2),
            ]
        ]

        self.assert_vars_equal("plev", *ds_list)

    def test_subset_level_by_values_and_bbox(self, tmpdir, mini_esgf_data):
        some_levels = [
            60000,
            50000,
            40000,
            30000,
            25000,
            20000,
            15000,
            10000,
            7000,
            5000,
        ]
        area = (20, 30.0, 150, 70.0)

        shuffled_1 = self.shuffle(some_levels)
        shuffled_2 = self.shuffle(some_levels)

        # Get various outputs and compare they are the same
        ds_list = [
            subset(
                ds=mini_esgf_data["CMIP6_TA"],
                output_dir=tmpdir,
                output_type="xarray",
                level=level,
                area=area,
            )[0]
            for level in [
                level_series(some_levels),
                level_interval(some_levels[0], some_levels[-1]),
                level_series(list(reversed(some_levels))),
                level_series(shuffled_1),
                level_series(shuffled_2),
            ]
        ]

        self.assert_vars_equal("plev", *ds_list)

    def test_subset_time_by_values_all(self, tmpdir, mini_esgf_data):
        all_times = [
            str(tm) for tm in xr.open_dataset(mini_esgf_data["CMIP6_TA"]).time.values
        ]

        shuffled_1 = self.shuffle(all_times)
        shuffled_2 = self.shuffle(all_times)

        # Get various outputs and compare they are the same
        ds_list = [
            subset(
                ds=mini_esgf_data["CMIP6_TA"],
                output_dir=tmpdir,
                output_type="xarray",
                time=times,
            )[0]
            for times in [
                None,
                time_series(all_times),
                time_interval(all_times[0], all_times[-1]),
                time_series(list(reversed(all_times))),
                time_series(shuffled_1),
                time_series(shuffled_2),
            ]
        ]

        self.assert_vars_equal("time", *ds_list)

    def test_subset_time_by_values_partial(self, tmpdir, mini_esgf_data):
        all_times = [
            str(tm) for tm in xr.open_dataset(mini_esgf_data["CMIP6_TA"]).time.values
        ]
        some_times = all_times[20:-15]

        shuffled_1 = self.shuffle(some_times)
        shuffled_2 = self.shuffle(some_times)

        # Get various outputs and compare they are the same
        ds_list = [
            subset(
                ds=mini_esgf_data["CMIP6_TA"],
                output_dir=tmpdir,
                output_type="xarray",
                time=times,
            )[0]
            for times in [
                time_interval(some_times[0], some_times[-1]),
                time_series(list(reversed(some_times))),
                time_series(shuffled_1),
                time_series(shuffled_2),
            ]
        ]

        self.assert_vars_equal("time", *ds_list)

    def test_subset_time_by_values_with_gaps(self, tmpdir, mini_esgf_data):
        all_times = [
            str(tm) for tm in xr.open_dataset(mini_esgf_data["CMIP6_TA"]).time.values
        ]
        some_times = [
            all_times[0],
            all_times[100],
            all_times[4],
            all_times[33],
            all_times[9],
        ]

        shuffled_1 = self.shuffle(some_times)
        shuffled_2 = self.shuffle(some_times)

        # Get various outputs and compare they are the same
        ds_list = [
            subset(
                ds=mini_esgf_data["CMIP6_TA"],
                output_dir=tmpdir,
                output_type="xarray",
                time=times,
            )[0]
            for times in [
                time_series(list(reversed(some_times))),
                time_series(shuffled_1),
                time_series(shuffled_2),
            ]
        ]

        self.assert_vars_equal("time", *ds_list)


def test_subset_by_time_components_year_month(tmpdir, mini_esgf_data):
    # times = ("2015-01-16 12", "MANY MORE", "2024-12-16 12") [120]
    tc1 = time_components(year=(2021, 2022), month=["dec", "jan", "feb"])
    tc2 = time_components(year=(2021, 2022), month=[12, 1, 2])

    kwargs = {"output_dir": tmpdir, "output_type": "xarray"}

    for tc in (tc1, tc2):
        with subset(mini_esgf_data["CMIP6_TA"], time_components=tc, **kwargs)[0] as ds:
            assert set(ds.time.dt.year.values) == {2021, 2022}
            assert set(ds.time.dt.month.values) == {12, 1, 2}


def test_subset_empty(tmpdir, mini_esgf_data):
    # Monthly mean dataset with 360-day calendar
    with pytest.raises(Exception) as exc:
        subset(
            ds=mini_esgf_data["CMIP5_TAS"],
            output_type="nc",
            output_dir=tmpdir,
            time_components="month:12,1,2|day:1,2,3,4,5",
        )
    assert str(exc.value) == "'No timesteps are matching the selection criteria.'"


def test_subset_by_time_components_31_days_360_day(tmpdir, mini_esgf_data):
    # Dataset with 360-day calendar
    tmpdir30 = Path(tmpdir, "ds30")
    tmpdir31 = Path(tmpdir, "ds31")
    os.mkdir(tmpdir30)
    os.mkdir(tmpdir31)

    subset(
        ds=mini_esgf_data["CMIP5_MRSOS_MULTIPLE_TIME_STEPS"],
        output_type="nc",
        output_dir=tmpdir30,
        time_components="month:12,1,2|day:1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30",
    )
    subset(
        ds=mini_esgf_data["CMIP5_MRSOS_MULTIPLE_TIME_STEPS"],
        output_type="nc",
        output_dir=tmpdir31,
        time_components="month:12,1,2|day:1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31",
    )
    with (
        xr.open_mfdataset(f"{tmpdir30}/*.nc") as ds30,
        xr.open_mfdataset(f"{tmpdir31}/*.nc") as ds31,
    ):
        assert ds30.identical(ds31)
        assert ds30.sizes["time"] == 8490


def test_subset_by_time_components_month_day(tmpdir, mini_esgf_data):
    # CMIP6_SICONC_DAY: 18500101-20141231 ;  n_times = 60225
    tc1 = time_components(month=["jul"], day=[1, 11, 21])
    tc2 = time_components(month=[7], day=[1, 11, 21])

    kwargs = {"output_dir": tmpdir, "output_type": "xarray"}

    for tc in (tc1, tc2):
        with subset(mini_esgf_data["CMIP6_SICONC_DAY"], time_components=tc, **kwargs)[
            0
        ] as ds:
            assert set(ds.time.dt.month.values) == {7}
            assert set(ds.time.dt.day.values) == {1, 11, 21}
            assert len(ds.time.values) == (2014 - 1850 + 1) * 3


def test_subset_by_time_interval_and_components_month_day(tmpdir, mini_esgf_data):
    # CMIP6_SICONC_DAY: 18500101-20141231 ;  n_times = 60225
    ys, ye = 1850, 1869
    ti = time_interval(f"{ys}-01-01T00:00:00", f"{ye}-12-31T23:59:59")

    months = [3, 4, 5]
    days = [5, 6]

    tc1 = time_components(month=["mar", "apr", "may"], day=days)
    tc2 = time_components(month=months, day=days)

    kwargs = {"output_dir": tmpdir, "output_type": "xarray"}

    for tc in (tc1, tc2):
        with subset(
            mini_esgf_data["CMIP6_SICONC_DAY"], time=ti, time_components=tc, **kwargs
        )[0] as ds:
            assert set(ds.time.dt.month.values) == set(months)
            assert set(ds.time.dt.day.values) == set(days)
            assert len(ds.time.values) == (ye - ys + 1) * len(months) * len(days)


def test_subset_by_time_series_and_components_month_day(tmpdir, mini_esgf_data):
    # CMIP6_SICONC_DAY: 18500101-20141231 ;  n_times = 60225
    ys, ye = 1850, 1869
    req_times = [
        tm.isoformat()
        for tm in xr.open_dataset(mini_esgf_data["CMIP6_SICONC_DAY"]).time.values
        if ys <= tm.year <= ye
    ]

    ts = time_series(req_times)
    months = [3, 4, 5]
    days = [5, 6]

    tc1 = time_components(month=["mar", "apr", "may"], day=days)
    tc2 = time_components(month=months, day=days)

    kwargs = {"output_dir": tmpdir, "output_type": "xarray"}

    for tc in (tc1, tc2):
        with subset(
            mini_esgf_data["CMIP6_SICONC_DAY"], time=ts, time_components=tc, **kwargs
        )[0] as ds:
            assert set(ds.time.dt.month.values) == set(months)
            assert set(ds.time.dt.day.values) == set(days)
            assert len(ds.time.values) == (ye - ys + 1) * len(months) * len(days)


def test_subset_by_area_and_components_month_day(tmpdir, mini_esgf_data):
    # CMIP6_SICONC_DAY: 18500101-20141231 ;  n_times = 60225
    ys, ye = 1850, 1869
    ti = time_interval(f"{ys}-01-01T00:00:00", f"{ye}-12-31T23:59:59")

    months = [3, 4, 5]
    days = [5, 6]

    tc1 = time_components(month=["mar", "apr", "may"], day=days)
    tc2 = time_components(month=months, day=days)

    area = area_parameter.AreaParameter((20, 30.0, 150, 70.0))

    kwargs = {"output_dir": tmpdir, "output_type": "xarray"}

    for tc in (tc1, tc2):
        with subset(
            mini_esgf_data["CMIP6_SICONC_DAY"],
            time=ti,
            time_components=tc,
            area=area,
            **kwargs,
        )[0] as ds:
            assert set(ds.time.dt.month.values) == set(months)
            assert set(ds.time.dt.day.values) == set(days)
            assert len(ds.time.values) == (ye - ys + 1) * len(months) * len(days)


@pytest.mark.thread_unsafe
@pytest.mark.slow
class TestSubsetResultLocking:

    def test_subset_nc_no_fill_value(self, nimbus, tmpdir, mini_esgf_data):
        """Tests clisops subset function with a time subset."""
        # FIXME: The `subset` operation relies on opening a file with xarray as a reference.
        # This can cause issues when modifications are done to the file from other processes.
        result = subset(
            ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
            time=time_interval("2005-01-01T00:00:00", "2020-12-30T00:00:00"),
            output_dir=tmpdir,
            output_type="nc",
            file_namer="simple",
        )

        # check that with just opening the file with xarray, saving to netcdf, then opening again, these fill values get added
        with _load_ds(mini_esgf_data["CMIP5_TAS"]) as ds:
            ds.to_netcdf(f"{tmpdir}/test_fill_values.nc")

        with _load_ds(f"{tmpdir}/test_fill_values.nc") as ds:
            # assert np.isnan(float(ds.time.encoding.get("_FillValue")))
            assert np.isnan(float(ds.lat.encoding.get("_FillValue")))
            assert np.isnan(float(ds.lon.encoding.get("_FillValue")))
            assert np.isnan(float(ds.height.encoding.get("_FillValue")))

            assert np.isnan(float(ds.lat_bnds.encoding.get("_FillValue")))
            assert np.isnan(float(ds.lon_bnds.encoding.get("_FillValue")))
            assert np.isnan(float(ds.time_bnds.encoding.get("_FillValue")))

        # FIXME: The `subset` operation relies on opening a file with xarray as a reference.
        # This can cause issues when modifications are done to the file from other processes.
        # check that there is no fill value in encoding for coordinate variables and bounds
        with _load_ds(result) as res:
            assert "_FillValue" not in res.time.encoding
            assert "_FillValue" not in res.lat.encoding
            assert "_FillValue" not in res.lon.encoding
            assert "_FillValue" not in res.height.encoding

            assert "_FillValue" not in res.lat_bnds.encoding
            assert "_FillValue" not in res.lon_bnds.encoding
            assert "_FillValue" not in res.time_bnds.encoding

    def test_subset_cmip5_nc_consistent_bounds(self, nimbus, tmpdir):
        """Tests clisops subset function with a time subset and check the metadata"""
        # FIXME: The `subset` operation relies on opening a file with xarray as a reference.
        # This can cause issues when modifications are done to the file from other processes.
        result = subset(
            ds=nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc"),
            time=time_interval("2005-01-01T00:00:00", "2020-12-30T00:00:00"),
            output_dir=tmpdir,
            output_type="nc",
            file_namer="simple",
        )
        with _load_ds(result) as res:
            # check fill value in bounds
            assert "_FillValue" not in res.lat_bnds.encoding
            assert "_FillValue" not in res.lon_bnds.encoding
            assert "_FillValue" not in res.time_bnds.encoding
            # check fill value in coordinates
            assert "_FillValue" not in res.time.encoding
            assert "_FillValue" not in res.lat.encoding
            assert "_FillValue" not in res.lon.encoding
            assert "_FillValue" not in res.height.encoding
            # check coordinates in bounds
            assert "coordinates" not in res.lat_bnds.encoding
            assert "coordinates" not in res.lon_bnds.encoding
            assert "coordinates" not in res.time_bnds.encoding

    def test_subset_cmip6_nc_consistent_bounds(self, nimbus, tmpdir, mini_esgf_data):
        """Tests clisops subset function with a time subset and check the metadata"""
        # FIXME: The `subset` operation relies on opening a file with xarray as a reference.
        # This can cause issues when modifications are done to the file from other processes.
        result = subset(
            ds=mini_esgf_data["CMIP6_TASMIN"],
            time=time_interval("2010-01-01T00:00:00", "2010-12-31T00:00:00"),
            output_dir=tmpdir,
            output_type="nc",
            file_namer="simple",
        )
        with _load_ds(result) as res:
            # check fill value in bounds
            assert "_FillValue" not in res.lat_bnds.encoding
            assert "_FillValue" not in res.lon_bnds.encoding
            assert "_FillValue" not in res.time_bnds.encoding
            # check fill value in coordinates
            assert "_FillValue" not in res.time.encoding
            assert "_FillValue" not in res.lat.encoding
            assert "_FillValue" not in res.lon.encoding
            assert "_FillValue" not in res.height.encoding
            # check coordinates in bounds
            assert "coordinates" not in res.lat_bnds.encoding
            assert "coordinates" not in res.lon_bnds.encoding
            assert "coordinates" not in res.time_bnds.encoding

    def test_subset_cmip6_issue_308_fillvalue(self, tmpdir, capsys, mini_esgf_data):
        """Tests clisops subset function with a time subset and check the metadata.

        Notes
        -----
        This test is used for fillvalue issues. See: https://github.com/roocs/clisops/issues/308
        """
        from clisops.utils.common import enable_logging
        from clisops.utils.testing import ContextLogger

        with ContextLogger():
            enable_logging()
            # FIXME: The `subset` operation relies on opening a file with xarray as a reference.
            # This can cause issues when modifications are done to the file from other processes.
            result = subset(
                ds=mini_esgf_data["CMIP6_FILLVALUE"],
                time=time_interval("2000-01-01T00:00:00", "2000-12-31T00:00:00"),
                output_dir=tmpdir,
                output_type="nc",
                file_namer="simple",
            )
            with _load_ds(result) as res:
                # check fill value in bounds
                assert "_FillValue" not in res.lat_bnds.encoding
                assert "_FillValue" not in res.lon_bnds.encoding
                assert "_FillValue" not in res.time_bnds.encoding
                # xarray should set the appropriate dtype
                assert res.tas.encoding["_FillValue"].dtype == np.float32
                assert res.tas.encoding["missing_value"].dtype == np.float32
                captured = capsys.readouterr()
                assert (
                    "The defined _FillValue and missing_value for 'tas' are not the same '1.0000000200408773e+20' != '1e+20'. Setting '1e+20' for both."
                    in captured.err
                )



import pytest

from clisops.exceptions import InvalidParameterValue
from clisops.parameter import AreaParameter, area

type_error = (
    "Input type of <{}> not allowed. Must be one of: "
    "[<class 'collections.abc.Sequence'>, <class 'str'>, <class 'clisops.parameter._utils.Series'>, <class 'NoneType'>]"
)


def test__str__():
    _area = "0.,49.,10.,65"
    parameter = AreaParameter(_area)
    assert parameter.__str__() == "Area to subset over:\n (0.0, 49.0, 10.0, 65.0)"
    assert parameter.__repr__() == parameter.__str__()
    assert parameter.__unicode__() == parameter.__str__()

    _area = area("0.,49.,10.,65")
    parameter = AreaParameter(_area)
    assert parameter.__str__() == "Area to subset over:\n (0.0, 49.0, 10.0, 65.0)"
    assert parameter.__repr__() == parameter.__str__()
    assert parameter.__unicode__() == parameter.__str__()


def test_raw():
    _area = area("0.,49.,10.,65")
    parameter = AreaParameter(_area)
    assert parameter.raw == _area


def test_tuple():
    _area = "0.,49.,10.,65"
    parameter = AreaParameter(_area)
    assert parameter.value == (0.0, 49.0, 10.0, 65)


def test_area_is_tuple_string():
    _area = ("0", "-10", "120", "40")
    parameter = AreaParameter(_area)
    assert parameter.value == (0.0, -10.0, 120.0, 40.0)


def test_input_list():
    _area = [0, 49.5, 10, 65]
    parameter = AreaParameter(_area)
    assert parameter.value == (0.0, 49.5, 10.0, 65)


def test_validate_error_number():
    _area = 0
    with pytest.raises(InvalidParameterValue) as exc:
        AreaParameter(_area)
    assert str(exc.value) == type_error.format("class 'int'")


def test_validate_error_words():
    _area = ["test", "_area", "error", "words"]
    with pytest.raises(InvalidParameterValue) as exc:
        AreaParameter(_area)
    assert str(exc.value) == "Values must be valid numbers"


def test_validate_error_len_1_tuple():
    _area = (0, 65)
    with pytest.raises(InvalidParameterValue) as exc:
        AreaParameter(_area)
    assert str(exc.value) == "AreaParameter should be of length 4 but is of length 2"


def test_asdict():
    _area = "0.,49.,10.,65"
    parameter = AreaParameter(_area)
    assert parameter.asdict() == {"lon_bnds": (0, 10), "lat_bnds": (49, 65)}


def test_whitespace():
    _area = "0., 49., 10., 65"
    parameter = AreaParameter(_area)
    assert parameter.value == (0.0, 49.0, 10.0, 65)


def test_empty_string():
    _area = ""
    assert AreaParameter(_area).asdict() is None
    assert AreaParameter(_area).value is None


def test_none():
    _area = None
    assert AreaParameter(_area).asdict() is None
    assert AreaParameter(_area).value is None


def test_none_2():
    _area = None
    assert AreaParameter(_area).asdict() is None
    assert AreaParameter(_area).value is None


def test_class_instance():
    _area = "0.,49.,10.,65"
    parameter = AreaParameter(_area)
    new_parameter = AreaParameter(parameter)
    assert new_parameter.value == (0.0, 49.0, 10.0, 65.0)



"""Test the different methods of transforming the longitude coordinates after rolling"""

import numpy as np


# calculate the offset in the same way as in dataset_utils
def calculate_offset(a, bounds):
    low, high = bounds
    # get resolution of data
    res = a[1] - a[0]

    # calculate how many degrees to move by to have lon[0] of rolled subset as lower bound of request
    diff = a[0] - low

    # work out how many elements to roll by to roll data by 1 degree
    index = 1 / res

    # calculate the corresponding offset needed to change data by diff
    # round up to ensure rolling by enough
    # offset = math.ceil(diff * index)
    offset = int(round(diff * index))

    return offset


def dataset_roll(a, offset, bounds):
    # roll the dataset
    low, high = bounds
    a_roll = np.roll(a, offset)

    if offset < 0:
        a_new = np.where(
            np.logical_and(low >= a_roll, a_roll <= -(360 + low)), a_roll, a_roll % 360
        )  # this doesn't work in all negative offset cases
    else:
        a_new = np.where(a_roll < (360 + low), a_roll, a_roll % -360)

    return a_new


def dataset_roll_using_offset(a, offset):
    # roll the dataset
    a_roll = np.roll(a, offset)

    if offset < 0:
        a_roll[offset:] = a_roll[offset:] % 360
    else:
        a_roll[:offset] = a_roll[:offset] % -360

    return a_roll


class TestLonRoll_0_360:
    # offset = 359
    def test_to_minus_359_0(self):
        a = np.arange(start=0, stop=360, step=1)
        bounds = (-359, 0)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-359, stop=1, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-359, stop=1, step=1))

    # offset = 270
    def test_to_minus_270_to_89(self):
        a = np.arange(start=0, stop=360, step=1)
        bounds = (-270, 89)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-270, stop=90, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-270, stop=90, step=1))

    # offset = 180
    def test_to_minus_180_179(self):
        a = np.arange(start=0, stop=360, step=1)
        bounds = (-180, 179)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-180, stop=180, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-180, stop=180, step=1))

    # offset = 90
    def test_to_minus_90_269(self):
        a = np.arange(start=0, stop=360, step=1)
        bounds = (-90, 269)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-90, stop=270, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-90, stop=270, step=1))

    # offset = 0
    def test_to_0_359(self):
        a = np.arange(start=0, stop=360, step=1)
        bounds = (0, 359)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=0, stop=360, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=0, stop=360, step=1))


class TestLonRoll_minus_90_270:
    # offset = 269
    def test_to_minus_359_0(self):
        a = np.arange(start=-90, stop=270, step=1)
        bounds = (-359, 0)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-359, stop=1, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-359, stop=1, step=1))

    # offset = 180
    def test_to_minus_270_to_89(self):
        a = np.arange(start=-90, stop=270, step=1)
        bounds = (-270, 89)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-270, stop=90, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-270, stop=90, step=1))

    # offset = 90
    def test_to_minus_180_179(self):
        a = np.arange(start=-90, stop=270, step=1)
        bounds = (-180, 179)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-180, stop=180, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-180, stop=180, step=1))

    # offset = 0
    def test_to_minus_90_269(self):
        a = np.arange(start=-90, stop=270, step=1)
        bounds = (-90, 269)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-90, stop=270, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-90, stop=270, step=1))

    # offset = -90 - FAILS
    def test_to_0_359(self):
        a = np.arange(start=-90, stop=270, step=1)
        bounds = (0, 359)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=0, stop=360, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=0, stop=360, step=1))


class TestLonRoll_minus_180_180:
    # offset = 179
    def test_to_minus_359_0(self):
        a = np.arange(start=-180, stop=180, step=1)
        bounds = (-359, 0)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-359, stop=1, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-359, stop=1, step=1))

    # offset = 90
    def test_to_minus_270_to_89(self):
        a = np.arange(start=-180, stop=180, step=1)
        bounds = (-270, 89)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-270, stop=90, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-270, stop=90, step=1))

    # offset = 0
    def test_to_minus_180_179(self):
        a = np.arange(start=-180, stop=180, step=1)
        bounds = (-180, 179)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-180, stop=180, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-180, stop=180, step=1))

    # offset = -90 - FAILS
    def test_to_minus_90_269(self):
        a = np.arange(start=-180, stop=180, step=1)
        bounds = (-90, 269)
        offset = calculate_offset(a, bounds)

        # a_where = dataset_roll(a, offset, bounds)
        # assert np.array_equal(a_where, np.arange(start=-90, stop=270, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-90, stop=270, step=1))

    # offset = -180 - FAILS
    def test_to_0_359(self):
        a = np.arange(start=-180, stop=180, step=1)
        bounds = (0, 359)
        offset = calculate_offset(a, bounds)

        # a_where = dataset_roll(a, offset, bounds)
        # assert np.array_equal(a_where, np.arange(start=0, stop=360, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=0, stop=360, step=1))


class TestLonRoll_minus_270_90:
    # offset = 89
    def test_to_minus_359_0(self):
        a = np.arange(start=-270, stop=90, step=1)
        bounds = (-359, 0)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-359, stop=1, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-359, stop=1, step=1))

    # offset = 0
    def test_to_minus_270_to_89(self):
        a = np.arange(start=-270, stop=90, step=1)
        bounds = (-270, 89)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-270, stop=90, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-270, stop=90, step=1))

    # offset = -90 - fAILS
    def test_to_minus_180_179(self):
        a = np.arange(start=-270, stop=90, step=1)
        bounds = (-180, 179)
        offset = calculate_offset(a, bounds)

        # a_where = dataset_roll(a, offset, bounds)
        # assert np.array_equal(a_where, np.arange(start=-180, stop=180, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-180, stop=180, step=1))

    # offset = -180 - FAILS
    def test_to_minus_90_269(self):
        a = np.arange(start=-270, stop=90, step=1)
        bounds = (-90, 269)
        offset = calculate_offset(a, bounds)

        # a_where = dataset_roll(a, offset, bounds)
        # assert np.array_equal(a_where, np.arange(start=-90, stop=270, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-90, stop=270, step=1))

    # offset = -270 - FAILS
    def test_to_0_359(self):
        a = np.arange(start=-270, stop=90, step=1)
        bounds = (0, 359)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=0, stop=360, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=0, stop=360, step=1))


class TestLonRoll_minus_360_0:
    # offset = 0
    def test_to_minus_359_0(self):
        a = np.arange(start=-359, stop=1, step=1)
        bounds = (-359, 0)
        offset = calculate_offset(a, bounds)

        a_where = dataset_roll(a, offset, bounds)
        assert np.array_equal(a_where, np.arange(start=-359, stop=1, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-359, stop=1, step=1))

    # offset = -89
    def test_to_minus_270_to_89(self):
        a = np.arange(start=-359, stop=1, step=1)
        bounds = (-270, 89)
        offset = calculate_offset(a, bounds)

        # a_where = dataset_roll(a, offset, bounds)
        # assert np.array_equal(a_where, np.arange(start=-270, stop=90, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-270, stop=90, step=1))

    # offset = -179
    def test_to_minus_180_179(self):
        a = np.arange(start=-359, stop=1, step=1)
        bounds = (-180, 179)
        offset = calculate_offset(a, bounds)

        # a_where = dataset_roll(a, offset, bounds)
        # assert np.array_equal(a_where, np.arange(start=-180, stop=180, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-180, stop=180, step=1))

    # offset = -269
    def test_to_minus_90_269(self):
        a = np.arange(start=-359, stop=1, step=1)
        bounds = (-90, 269)
        offset = calculate_offset(a, bounds)

        # a_where = dataset_roll(a, offset, bounds)
        # assert np.array_equal(a_where, np.arange(start=-90, stop=270, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=-90, stop=270, step=1))

    # offset = -359
    def test_to_0_359(self):
        a = np.arange(start=-359, stop=1, step=1)
        bounds = (0, 359)
        offset = calculate_offset(a, bounds)

        # a_where = dataset_roll(a, offset, bounds)
        # assert np.array_equal(a_where, np.arange(start=0, stop=360, step=1))

        a_offset = dataset_roll_using_offset(a, offset)
        assert np.array_equal(a_offset, np.arange(start=0, stop=360, step=1))



from clisops.exceptions import InvalidParameterValue
from clisops.parameter._utils import interval, series


class _BaseParameter:
    """Base class for parameters used in operations (e.g. subset, average etc.)."""

    allowed_input_types = None

    def __init__(self, input):
        self.input = self.raw = input

        # If the input is already an instance of this class, call its parse method
        if isinstance(self.input, self.__class__):
            self.value = self.input.value
            self.type = getattr(self.input, "type", "undefined")
        else:
            self._check_input_type()
            self.value = self._parse()

    def _check_input_type(self):
        if not self.allowed_input_types:
            return
        if not isinstance(self.input, tuple(self.allowed_input_types)):
            raise InvalidParameterValue(
                f"Input type of {type(self.input)} not allowed. "
                f"Must be one of: {self.allowed_input_types}"
            )

    def _parse(self):
        raise NotImplementedError()

    def get_bounds(self):
        """Returns a tuple of the (start, end) times, calculated from
        the value of the parameter. Either will default to None."""
        raise NotImplementedError()

    def __str__(self):
        raise NotImplementedError()

    def __repr__(self):
        return str(self)

    def __unicode__(self):
        return str(self)


class _BaseIntervalOrSeriesParameter(_BaseParameter):
    """
    A base class for a parameter that can be instantiated from either and `Interval` or `Series` class instance.

    It has a `type` and a `value` reflecting the type, e.g.:
        - type: "interval" --> value: (start, end)
        - type: "series"   --> value: [item1, item2, ..., item_n]
    """

    allowed_input_types = [interval, series, type(None), str]

    def _parse(self):
        if isinstance(self.input, interval):
            self.type = "interval"
            return self._parse_as_interval()
        elif isinstance(self.input, series):
            self.type = "series"
            return self._parse_as_series()
        elif isinstance(self.input, type(None)):
            self.type = "none"
            return None
        elif isinstance(self.input, str):
            if "/" in self.input:
                self.type = "interval"
                self.input = interval(self.input)
                return self._parse_as_interval()
            else:
                self.type = "series"
                self.input = series(self.input)
                return self._parse_as_series()

    def _parse_as_interval(self):
        raise NotImplementedError()

    def _parse_as_series(self):
        raise NotImplementedError()

    def _value_as_tuple(self):
        value = self.value
        if value is None:
            value = None, None

        return value



from collections.abc import Sequence

from clisops.exceptions import InvalidParameterValue
from clisops.parameter._utils import dimensions, parse_sequence
from clisops.parameter.base_parameter import _BaseParameter
from clisops.utils.dataset_utils import known_coord_types


class DimensionParameter(_BaseParameter):
    """
    Class for dimensions parameter used in averaging operation.

    | Area can be input as:
    | A string of comma separated values: "time,latitude,longitude"
    | A sequence of strings: ("time", "longitude")

    Dimensions can be None or any number of options from time, latitude, longitude and level provided these
    exist in the dataset being operated on.

    Validates the dims input and parses the values into a sequence of strings.

    """

    allowed_input_types = [Sequence, str, dimensions, type(None)]

    def _parse(self):
        classname = self.__class__.__name__

        if self.input in (None, ""):
            return None
        elif isinstance(self.input, dimensions):
            value = self.input.value
        else:
            value = parse_sequence(self.input, caller=classname)

        for item in value:
            if not isinstance(item, str):
                raise InvalidParameterValue("Each dimension must be a string.")

            if item not in known_coord_types:
                raise InvalidParameterValue(
                    f"Dimensions for averaging must be one of {known_coord_types}"
                )

        return tuple(value)

    def asdict(self):
        """Returns a dictionary of the dimensions"""
        if self.value is not None:
            return {"dims": self.value}

    def __str__(self):
        return f"Dimensions to average over:" f"\n {self.value}"



import glob

import cf_xarray as cfxr  # noqa
import numpy as np
import packaging.version
import pytest
import xarray as xr
from packaging.version import Version
from roocs_grids import get_grid_file

import clisops.utils.dataset_utils as clidu
from clisops.core.regrid import XESMF_MINIMUM_VERSION

try:
    import xesmf

    if Version(xesmf.__version__) < Version(XESMF_MINIMUM_VERSION):
        raise ImportError("xESMF version is too low.")
except ImportError:
    xesmf = None

XESMF_IMPORT_MSG = (
    f"xESMF >= {XESMF_MINIMUM_VERSION} is needed for regridding functionalities."
)


def test_add_day(mini_esgf_data):
    with xr.open_dataset(
        mini_esgf_data["CMIP6_SICONC"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    ) as ds:
        date = "2012-02-29T00:00:00"

        new_date = clidu.adjust_date_to_calendar(ds, date, "forwards")

        assert new_date == "2012-03-01T00:00:00"


def test_sub_day(mini_esgf_data):
    with xr.open_dataset(
        mini_esgf_data["CMIP6_SICONC"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    ) as ds:
        date = "2012-02-30T00:00:00"

        new_date = clidu.adjust_date_to_calendar(ds, date, "backwards")

        assert new_date == "2012-02-28T00:00:00"


def test_invalid_day(mini_esgf_data):
    with xr.open_dataset(
        mini_esgf_data["CMIP6_SICONC"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    ) as ds:
        date = "2012-02-29T00:00:00"

        with pytest.raises(Exception) as exc:
            clidu.adjust_date_to_calendar(ds, date, "odd")
        assert (
            str(exc.value)
            == "Invalid value for direction: odd. This should be either 'backwards' to indicate subtracting a day or 'forwards' for adding a day."
        )


def test_date_out_of_expected_range(mini_esgf_data):
    da = xr.open_dataset(
        mini_esgf_data["CMIP6_SICONC"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    )
    date = "2012-00-01T00:00:00"

    with pytest.raises(Exception) as exc:
        clidu.adjust_date_to_calendar(da, date, "forwards")
    assert (
        str(exc.value) == "Invalid input 0 for month. Expected value between 1 and 12."
    )


def test_add_hor_CF_coord_attrs():
    """Test function to add standard attributes to horizontal coordinate variables."""
    # Create basic dataset
    with xr.Dataset(
        data_vars={},
        coords={
            "lat": (["lat"], np.ones(1)),
            "lon": (["lon"], np.ones(1)),
            "lat_bnds": (["lat", "bnds"], np.ones((1, 2))),
            "lon_bnds": (["lon", "bnds"], np.ones((1, 2))),
        },
    ) as ds:

        # Ensuring attributes have been added
        ds = clidu.add_hor_CF_coord_attrs(ds=ds)
        assert ds["lat"].attrs["bounds"] == "lat_bnds"
        assert ds["lon"].attrs["bounds"] == "lon_bnds"
        assert ds["lat"].attrs["units"] == "degrees_north"
        assert ds["lon"].attrs["units"] == "degrees_east"
        assert ds["lat"].attrs["axis"] == "Y"
        assert ds["lon"].attrs["axis"] == "X"
        assert ds["lat"].attrs["standard_name"] == "latitude"
        assert ds["lon"].attrs["standard_name"] == "longitude"

        # Ensuring attributes have been updated (and conflicting ones overwritten)
        ds["lat"].attrs["bounds"] = "lat_b"
        ds["lon"].attrs["standard_name"] = "lon"
        ds["lat_bnds"].attrs["custom"] = "custom"
        ds = clidu.add_hor_CF_coord_attrs(ds=ds, keep_attrs=True)
        assert ds["lat"].attrs["bounds"] == "lat_bnds"
        assert ds["lon"].attrs["bounds"] == "lon_bnds"
        assert ds["lat"].attrs["units"] == "degrees_north"
        assert ds["lon"].attrs["units"] == "degrees_east"
        assert ds["lat"].attrs["axis"] == "Y"
        assert ds["lon"].attrs["axis"] == "X"
        assert ds["lat"].attrs["standard_name"] == "latitude"
        assert ds["lon"].attrs["standard_name"] == "longitude"
        assert ds["lat_bnds"].attrs["custom"] == "custom"

        # Incorrect coordinate variable name supplied should lead to a KeyError
        with pytest.raises(KeyError) as exc:
            ds = clidu.add_hor_CF_coord_attrs(ds, lat="latitude")
        assert (
            str(exc.value)
            == "'Not all specified coordinate variables exist in the dataset.'"
        )


@pytest.mark.skipif(xesmf is None, reason=XESMF_IMPORT_MSG)
def test_reformat_xESMF_to_CF():
    """Test reformat operation reformat_xESMF_to_CF."""
    # Use xesmf utility function to create dataset with global grid
    with xesmf.util.grid_global(5.0, 5.0) as ds:

        # It should have certain variables defined
        assert all([coord in ds for coord in ["lat", "lon", "lat_b", "lon_b"]])
        assert all([dim in ds.dims for dim in ["x", "y", "x_b", "y_b"]])

        # Reformat
        ds.attrs["xesmf"] = xesmf.__version__
        ds_ref = clidu.reformat_xESMF_to_CF(ds=ds, keep_attrs=True)
        assert all(
            [coord in ds_ref for coord in ["lat", "lon", "lat_bnds", "lon_bnds"]]
        )
        assert all([dim in ds_ref.dims for dim in ["lat", "lon", "bnds"]])
        assert ds_ref.attrs["xesmf"] == xesmf.__version__


def test_reformat_SCRIP_to_CF():
    """Test reformat operation reformat_SCRIP_to_CF."""
    # Load dataset in SCRIP format (using roocs_grids)
    with xr.open_dataset(get_grid_file("2pt5deg")) as ds:

        # It should have certain variables defined
        assert all(
            [
                coord in ds
                for coord in [
                    "grid_center_lat",
                    "grid_center_lon",
                    "grid_corner_lat",
                    "grid_corner_lon",
                    "grid_dims",
                    "grid_area",
                    "grid_imask",
                ]
            ]
        )
        assert all(
            [dim in ds.dims for dim in ["grid_corners", "grid_size", "grid_rank"]]
        )

        # Reformat
        ds_ref = clidu.reformat_SCRIP_to_CF(ds=ds, keep_attrs=True)
        assert all(
            [coord in ds_ref for coord in ["lat", "lon", "lat_bnds", "lon_bnds"]]
        )
        assert all([dim in ds_ref.dims for dim in ["lat", "lon", "bnds"]])
        assert ds_ref.attrs["Conventions"] == "SCRIP"


def test_detect_shape_regular():
    """Test detect_shape function for a regular grid."""
    # Load dataset
    with xr.open_dataset(get_grid_file("0pt25deg_era5_lsm")) as ds:

        # Detect shape
        nlat, nlon, ncells = clidu.detect_shape(
            ds, lat="latitude", lon="longitude", grid_type="regular_lat_lon"
        )

        # Assertion
        assert nlat == 721
        assert nlon == 1440
        assert ncells == nlat * nlon


def test_detect_shape_unstructured(mini_esgf_data):
    """Test detect_shape function for an unstructured grid."""
    # Load dataset
    with xr.open_dataset(
        mini_esgf_data["CMIP6_UNSTR_ICON_A"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    ) as ds:

        # Detect shape
        nlat, nlon, ncells = clidu.detect_shape(
            ds, lat="latitude", lon="longitude", grid_type="unstructured"
        )

        # Assertion
        assert nlat == ncells
        assert nlon == ncells
        assert ncells == 20480


@pytest.mark.skipif(xesmf is None, reason=XESMF_IMPORT_MSG)
def test_detect_format():
    """Test detect_format function."""
    # Load/Create datasets in SCRIP, CF and xESMF format
    with (
        xr.open_dataset(get_grid_file("0pt25deg_era5_lsm")) as ds_cf,
        xr.open_dataset(get_grid_file("0pt25deg_era5")) as ds_scrip,
        xesmf.util.grid_global(5.0, 5.0) as ds_xesmf,
    ):

        # Assertion
        assert clidu.detect_format(ds_cf) == "CF"
        assert clidu.detect_format(ds_scrip) == "SCRIP"
        assert clidu.detect_format(ds_xesmf) == "xESMF"


def test_detect_coordinate_and_bounds(mini_esgf_data):
    """Test detect_bounds and detect_coordinate functions."""
    ds_a = xr.open_mfdataset(
        mini_esgf_data["C3S_CORDEX_AFR_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ).load()
    ds_b = xr.open_mfdataset(
        mini_esgf_data["C3S_CORDEX_ANT_SFC_WIND"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ).load()
    ds_c = xr.open_dataset(mini_esgf_data["CMIP6_UNSTR_ICON_A"]).load()
    ds_d = xr.open_dataset(mini_esgf_data["CMIP6_OCE_HALO_CNRM"]).load()

    # check lat, lon are found
    lat_a = clidu.detect_coordinate(ds_a, "latitude")
    lon_a = clidu.detect_coordinate(ds_a, "longitude")
    lat_b = clidu.detect_coordinate(ds_b, "latitude")
    lon_b = clidu.detect_coordinate(ds_b, "longitude")
    lat_c = clidu.detect_coordinate(ds_c, "latitude")
    lon_c = clidu.detect_coordinate(ds_c, "longitude")
    lat_d = clidu.detect_coordinate(ds_d, "latitude")
    lon_d = clidu.detect_coordinate(ds_d, "longitude")

    # assert the correct variables have been found
    assert lat_a == "lat"
    assert lon_a == "lon"
    assert lat_b == "lat"
    assert lon_b == "lon"
    assert lat_c == "latitude"
    assert lon_c == "longitude"
    assert lat_d == "lat"
    assert lon_d == "lon"

    # assert detected bounds
    assert clidu.detect_bounds(ds_a, lat_a) == "lat_vertices"
    assert clidu.detect_bounds(ds_a, lon_a) == "lon_vertices"
    assert clidu.detect_bounds(ds_b, lat_b) is None
    assert clidu.detect_bounds(ds_b, lon_b) is None
    assert clidu.detect_bounds(ds_c, lat_c) == "latitude_bnds"
    assert clidu.detect_bounds(ds_c, lon_c) == "longitude_bnds"
    assert clidu.detect_bounds(ds_d, lat_d) == "lat_bnds"
    assert clidu.detect_bounds(ds_d, lon_d) == "lon_bnds"

    # test that latitude and longitude are still found when they are data variables
    # reset coords sets lat and lon as data variables
    ds_a = ds_a.reset_coords([lat_a, lon_a])
    ds_b = ds_b.reset_coords([lat_b, lon_b])
    ds_c = ds_c.reset_coords([lat_c, lon_c])
    ds_d = ds_d.reset_coords([lat_d, lon_d])
    assert lat_a == clidu.detect_coordinate(ds_a, "latitude")
    assert lon_a == clidu.detect_coordinate(ds_a, "longitude")
    assert lat_b == clidu.detect_coordinate(ds_b, "latitude")
    assert lon_b == clidu.detect_coordinate(ds_b, "longitude")
    assert lat_c == clidu.detect_coordinate(ds_c, "latitude")
    assert lon_c == clidu.detect_coordinate(ds_c, "longitude")
    assert lat_d == clidu.detect_coordinate(ds_d, "latitude")
    assert lon_d == clidu.detect_coordinate(ds_d, "longitude")


def test_detect_coordinate_robustness(tmpdir, mini_esgf_data):
    """Test coordinate detection for dataset where cf_xarray fails due to erroneous attributes."""
    with xr.open_mfdataset(
        mini_esgf_data["C3S_CORDEX_AFR_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ).load() as ds:
        assert clidu.detect_coordinate(ds, "latitude") == "lat"
        assert clidu.detect_coordinate(ds, "longitude") == "lon"

        # Set illegal units attribute to confuse cf_xarray
        ds.rlat.attrs["units"] = "degrees_north"
        ds.rlon.attrs["units"] = "degrees_east"
        with pytest.raises(
            KeyError, match="Receive multiple variables for key 'latitude':"
        ):
            ds.cf["latitude"]
        with pytest.raises(
            KeyError, match="Receive multiple variables for key 'longitude':"
        ):
            ds.cf["longitude"]
        assert clidu.detect_coordinate(ds, "latitude") == "lat"
        assert clidu.detect_coordinate(ds, "longitude") == "lon"

        # Additionally remove standard_name
        del ds.lat.attrs["standard_name"]
        del ds.lon.attrs["standard_name"]
        assert clidu.detect_coordinate(ds, "latitude") == "lat"
        assert clidu.detect_coordinate(ds, "longitude") == "lon"


def test_detect_gridtype(mini_esgf_data):
    """Test the function detect_gridtype."""
    ds_a = xr.open_dataset(
        mini_esgf_data["CMIP6_UNSTR_ICON_A"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    )
    ds_b = xr.open_dataset(
        mini_esgf_data["CMIP6_TOS_ONE_TIME_STEP"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    )
    ds_c = xr.open_dataset(
        mini_esgf_data["CMIP6_TAS_ONE_TIME_STEP"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    )
    assert (
        clidu.detect_gridtype(
            ds_a,
            lat="latitude",
            lon="longitude",
            lat_bnds="latitude_bnds",
            lon_bnds="longitude_bnds",
        )
        == "unstructured"
    )
    assert (
        clidu.detect_gridtype(
            ds_b,
            lat="latitude",
            lon="longitude",
            lat_bnds="vertices_latitude",
            lon_bnds="vertices_longitude",
        )
        == "curvilinear"
    )
    assert (
        clidu.detect_gridtype(
            ds_c, lat="lat", lon="lon", lat_bnds="lat_bnds", lon_bnds="lon_bnds"
        )
        == "regular_lat_lon"
    )


def test_determine_lon_lat_range(mini_esgf_data):
    """Test the function determine_lon_lat_range incl. and without fix for unmasked missing values."""
    # Test case 1 - without fix
    with xr.open_dataset(mini_esgf_data["CMIP6_EXTENT_UNMASKED"]) as ds:
        x0, x1, y0, y1 = clidu.determine_lon_lat_range(
            ds, "lon", "lat", "lon_bnds", "lat_bnds", apply_fix=False
        )
        for i, j in [
            (x0, -300),
            (x1, 1e20),
            (y0, -78),
            (y1, 1e20),
        ]:
            assert np.isclose(i, j, atol=0.1)

        # Test case 1 - with fix
        x0, x1, y0, y1 = clidu.determine_lon_lat_range(
            ds, "lon", "lat", "lon_bnds", "lat_bnds", apply_fix=True
        )
        for i, j in [
            (x0, -300.0),
            (x1, 60.0),
            (y0, -78),
            (y1, 89.9),
        ]:
            assert np.isclose(i, j, atol=0.1)

    # Test case 2 - without fix
    with xr.open_dataset(mini_esgf_data["CMIP6_UNTAGGED_MISSVALS"]) as ds:
        x0, x1, y0, y1 = clidu.determine_lon_lat_range(
            ds, "lon", "lat", "lon_bnds", "lat_bnds", apply_fix=False
        )
        for i, j in [
            (x0, 0.0),
            (x1, 9.9692e36),
            (y0, -79.2),
            (y1, 9.9692e36),
        ]:
            assert np.isclose(i, j, atol=0.1)

        # Test case 2 - with fix
        x0, x1, y0, y1 = clidu.determine_lon_lat_range(
            ds, "lon", "lat", "lon_bnds", "lat_bnds", apply_fix=True
        )
        for i, j in [
            (x0, 0.0),
            (x1, 360),
            (y0, -79.2),
            (y1, 89.7),
        ]:
            assert np.isclose(i, j, atol=0.1)

        # Fix is applied on the ds "in place"
        x0, x1, y0, y1 = clidu.determine_lon_lat_range(
            ds, "lon", "lat", "lon_bnds", "lat_bnds", apply_fix=False
        )
        for i, j in [
            (x0, 0.0),
            (x1, 360),
            (y0, -79.2),
            (y1, 89.7),
        ]:
            assert np.isclose(i, j, atol=0.1)


def test_determine_lon_lat_range_unstructured(mini_esgf_data):
    """Test the function determine_lon_lat_range for unstructured grids."""
    with xr.open_dataset(mini_esgf_data["CMIP6_UNSTR_ICON_A"]) as ds:
        # Test case 1 - manipulate only latitudes
        ds.latitude.values[0] = -999.0
        with pytest.warns(
            UserWarning,
            match="fix is not possible since their locations are not consistent",
        ):
            x0, x1, y0, y1 = clidu.determine_lon_lat_range(
                ds,
                "longitude",
                "latitude",
                "longitude_bnds",
                "latitude_bnds",
                apply_fix=True,
            )
        for i, j in [
            (x0, 1.0),
            (x1, 360.0),
            (y0, -999.0),
            (y1, 88.9),
        ]:
            assert np.isclose(i, j, atol=0.1)
        # Test case 2 - manipulate longitudes as well, but inconsistent
        ds.longitude.values[1] = -999.0
        with pytest.warns(
            UserWarning,
            match="fix is not possible since their locations are not consistent",
        ):
            x0, x1, y0, y1 = clidu.determine_lon_lat_range(
                ds,
                "longitude",
                "latitude",
                "longitude_bnds",
                "latitude_bnds",
                apply_fix=True,
            )
        for i, j in [
            (x0, -999.0),
            (x1, 360.0),
            (y0, -999.0),
            (y1, 88.9),
        ]:
            assert np.isclose(i, j, atol=0.1)
        # Test case 3 - manipulate latitudes and longitudes consistently
        ds.longitude.values[0] = -999.0
        ds.latitude.values[1] = -999.0
        with pytest.warns(UserWarning, match=r"missing_value found \(and treated\)"):
            x0, x1, y0, y1 = clidu.determine_lon_lat_range(
                ds,
                "longitude",
                "latitude",
                "longitude_bnds",
                "latitude_bnds",
                apply_fix=True,
            )
        for i, j in [
            (x0, 1.0),
            (x1, 360.0),
            (y0, -88.9),
            (y1, 88.9),
        ]:
            assert np.isclose(i, j, atol=0.1)


def test_determine_lon_lat_range_regular_lat_lon(mini_esgf_data):
    """Test the function determine_lon_lat_range for regular lat lon grids."""
    with xr.open_mfdataset(mini_esgf_data["CMIP5_TAS"]) as ds:
        ds.lat.values[1] = -999.0
        with pytest.warns(UserWarning, match="fix is not possible for regular"):
            clidu.determine_lon_lat_range(
                ds, "lon", "lat", "lon_bnds", "lat_bnds", apply_fix=True
            )


def test_crosses_0_meridian(mini_esgf_data):
    """Test the _crosses_0_meridian function."""
    # Case 1 - longitude crossing 180° meridian
    lon = np.arange(160.0, 200.0, 1.0)

    # convert to [-180, 180], min and max now suggest 0-meridian crossing
    lon = np.where(lon > 180, lon - 360, lon)

    da = xr.DataArray(dims=["x"], coords={"x": lon})
    assert not clidu._crosses_0_meridian(da["x"])

    # Case 2 - regional dataset ranging [315 .. 66] but for whatever reason not defined on
    #          [-180, 180] longitude frame
    ds = xr.open_dataset(mini_esgf_data["CORDEX_TAS_ONE_TIMESTEP"])
    assert np.isclose(ds["lon"].min(), 0, atol=0.1)
    assert np.isclose(ds["lon"].max(), 360, atol=0.1)

    # Convert to -180, 180 frame and confirm crossing 0-meridian
    ds, ll, lu = clidu.cf_convert_between_lon_frames(ds, (-180, 180))
    assert np.isclose(ds["lon"].min(), -45.4, atol=0.1)
    assert np.isclose(ds["lon"].max(), 66.1, atol=0.1)
    assert clidu._crosses_0_meridian(ds["lon"])


def test_convert_interval_between_lon_frames():
    """Test the helper function _convert_interval_between_lon_frames."""
    # Convert from 0,360 to -180,180 longitude frame and vice versa
    assert clidu._convert_interval_between_lon_frames(20, 60) == (20, 60)
    assert clidu._convert_interval_between_lon_frames(190, 200) == (-170, -160)
    assert clidu._convert_interval_between_lon_frames(-20, -90) == (270, 340)

    # Exception when crossing 0°- or 180°-meridian
    with pytest.raises(
        Exception,
        match="Cannot convert longitude interval if it includes the 0°- or 180°-meridian.",
    ):
        clidu._convert_interval_between_lon_frames(170, 300)
    with pytest.raises(
        Exception,
        match="Cannot convert longitude interval if it includes the 0°- or 180°-meridian.",
    ):
        clidu._convert_interval_between_lon_frames(-30, 10)


def test_convert_lon_frame_bounds():
    """Test the function cf_convert_between_lon_frames."""
    # Load tutorial dataset defined on [200,330]
    with xr.tutorial.open_dataset("air_temperature") as ds:
        assert ds["lon"].min() == 200.0
        assert ds["lon"].max() == 330.0

        # Create bounds
        dsb = ds.cf.add_bounds("lon")

        # Convert to other lon frame
        conv, ll, lu = clidu.cf_convert_between_lon_frames(dsb, (-180, 180))

        assert conv["lon"].values[0] == -160.0
        assert conv["lon"].values[-1] == -30.0

        # Check that bounds contain the respective cell centers
        if packaging.version.parse(xr.__version__) >= packaging.version.parse(
            "2022.11.0"
        ):
            assert np.all(
                conv["lon"].values[:] > conv["lon_bounds"].values[:, 0]
            )  # xarray >=2022.11.0 reorders dim values
            assert np.all(
                conv["lon"].values[:] < conv["lon_bounds"].values[:, 1]
            )  # xarray >=2022.11.0 reorders dim values
        else:
            assert np.all(conv["lon"].values[:] > conv["lon_bounds"].values[0, :])
            assert np.all(conv["lon"].values[:] < conv["lon_bounds"].values[1, :])

        # Convert only lon_interval
        conv, ll, lu = clidu.cf_convert_between_lon_frames(dsb, (-180, -10))

        assert conv["lon"].min() == 200.0
        assert conv["lon"].max() == 330.0
        assert ll == 180.0
        assert lu == 350.0


def test_convert_lon_frame_force():
    """Test the force option of function cf_convert_between_lon_frames"""
    # Load tutorial dataset defined on [200,330]
    ds = xr.tutorial.open_dataset("air_temperature")
    assert ds["lon"].min().item() == 200.0
    assert ds["lon"].max().item() == 330.0

    # Convert to other lon frame
    conv, ll, lu = clidu.cf_convert_between_lon_frames(ds, (-180, 180))

    assert conv["lon"].values[0] == -160.0
    assert conv["lon"].values[-1] == -30.0

    # Convert only lon_interval
    conv, ll, lu = clidu.cf_convert_between_lon_frames(ds, (-180, -10))

    assert conv["lon"].min().item() == 200.0
    assert conv["lon"].max().item() == 330.0
    assert ll == 180.0
    assert lu == 350.0

    # The same, but forcing the conversion of the longitudes
    conv, ll, lu = clidu.cf_convert_between_lon_frames(ds, (-180, -10), force=True)
    assert conv["lon"].min().item() == -160.0
    assert conv["lon"].max().item() == -30.0
    assert ll == -180.0
    assert lu == -10.0


def test_convert_lon_frame_shifted_bounds(mini_esgf_data):
    with xr.open_dataset(
        mini_esgf_data["CMIP6_GFDL_EXTENT"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    ) as ds:

        # confirm shifted frame
        assert np.isclose(ds["lon"].min(), -300.0, atol=0.5)
        assert np.isclose(ds["lon"].max(), 60.0, atol=0.5)

        # convert to [-180, 180]
        ds_a, ll, lu = clidu.cf_convert_between_lon_frames(ds, (-180, 180))
        assert (ll, lu) == (-180, 180)
        assert np.isclose(ds_a["lon"].min(), -180.0, atol=0.5)
        assert np.isclose(ds_a["lon"].max(), 180.0, atol=0.5)
        assert np.isclose(ds_a["lon_bnds"].min(), -180.0, atol=0.5)
        assert np.isclose(ds_a["lon_bnds"].max(), 180.0, atol=0.5)

        # convert to [0, 360]
        ds_b, ll, lu = clidu.cf_convert_between_lon_frames(ds, (0, 360))
        assert (ll, lu) == (0, 360)
        assert np.isclose(ds_b["lon"].min(), 0.0, atol=0.5)
        assert np.isclose(ds_b["lon"].max(), 360.0, atol=0.5)
        assert np.isclose(ds_b["lon_bnds"].min(), 0.0, atol=0.5)
        assert np.isclose(ds_b["lon_bnds"].max(), 360.0, atol=0.5)

        # convert intermediate result to [0, 360]
        ds_c, ll, lu = clidu.cf_convert_between_lon_frames(ds_a, (0, 360))
        assert (ll, lu) == (0, 360)
        assert np.isclose(ds_c["lon"].min(), 0.0, atol=0.5)
        assert np.isclose(ds_c["lon"].max(), 360.0, atol=0.5)
        assert np.isclose(ds_c["lon_bnds"].min(), 0.0, atol=0.5)
        assert np.isclose(ds_c["lon_bnds"].max(), 360.0, atol=0.5)

        # convert intermediate result to [-180, 180]
        ds_d, ll, lu = clidu.cf_convert_between_lon_frames(ds_a, (-180, 180))
        assert (ll, lu) == (-180, 180)
        assert np.isclose(ds_d["lon"].min(), -180.0, atol=0.5)
        assert np.isclose(ds_d["lon"].max(), 180.0, atol=0.5)
        assert np.isclose(ds_d["lon_bnds"].min(), -180.0, atol=0.5)
        assert np.isclose(ds_d["lon_bnds"].max(), 180.0, atol=0.5)

        # assert projection coordinate sorted
        assert np.all(ds_d["x"].values[1:] - ds_d["x"].values[:-1] > 0.0)
        assert np.all(ds_c["x"].values[1:] - ds_c["x"].values[:-1] > 0.0)


def test_convert_lon_frame_shifted_no_bounds(mini_esgf_data):
    with xr.open_dataset(
        mini_esgf_data["CMIP6_IITM_EXTENT"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    ) as ds:

        # confirm shifted frame
        assert np.isclose(ds["longitude"].min(), -280.0, atol=1.0)
        assert np.isclose(ds["longitude"].max(), 80.0, atol=1.0)

        # convert to [-180, 180]
        ds_a, ll, lu = clidu.cf_convert_between_lon_frames(ds, (-180, 180))
        assert (ll, lu) == (-180, 180)
        assert np.isclose(ds_a["longitude"].min(), -180.0, atol=1.0)
        assert np.isclose(ds_a["longitude"].max(), 180.0, atol=1.0)

        # convert to [0, 360]
        ds_b, ll, lu = clidu.cf_convert_between_lon_frames(ds, (0, 360))
        assert (ll, lu) == (0, 360)
        assert np.isclose(ds_b["longitude"].min(), 0.0, atol=1.0)
        assert np.isclose(ds_b["longitude"].max(), 360.0, atol=1.0)

        # convert intermediate result to [0, 360]
        ds_c, ll, lu = clidu.cf_convert_between_lon_frames(ds_a, (0, 360))
        assert (ll, lu) == (0, 360)
        assert np.isclose(ds_c["longitude"].min(), 0.0, atol=1.0)
        assert np.isclose(ds_c["longitude"].max(), 360.0, atol=1.0)

        # convert intermediate result to [-180, 180]
        ds_d, ll, lu = clidu.cf_convert_between_lon_frames(ds_a, (-180, 180))
        assert (ll, lu) == (-180, 180)
        assert np.isclose(ds_d["longitude"].min(), -180.0, atol=1.0)
        assert np.isclose(ds_d["longitude"].max(), 180.0, atol=1.0)

        # assert projection coordinate sorted
        assert np.all(ds_d["x"].values[1:] - ds_d["x"].values[:-1] > 0.0)
        assert np.all(ds_c["x"].values[1:] - ds_c["x"].values[:-1] > 0.0)


# todo: add a few more tests of cf_convert_lon_frame using xe.util functions to create regional and global datasets


def test_calculate_bounds_curvilinear(mini_esgf_data):
    """Test for bounds calculation for curvilinear grid."""

    # Load CORDEX dataset (curvilinear grid)
    with xr.open_dataset(mini_esgf_data["CORDEX_TAS_ONE_TIMESTEP"]).isel(
        {"rlat": range(100, 120), "rlon": range(100, 120)}
    ) as ds:

        # Drop bounds variables and compute them
        ds_nb = ds.drop_vars(["lon_vertices", "lat_vertices"])
        ds_nb = clidu.generate_bounds_curvilinear(ds_nb, lat="lat", lon="lon")

        # Sort every cells vertices values
        for i in range(1, 19):
            for j in range(1, 19):
                ds.lat_vertices[i, j, :] = np.sort(ds.lat_vertices.values[i, j, :])
                ds.lon_vertices[i, j, :] = np.sort(ds.lon_vertices.values[i, j, :])
                ds_nb.vertices_latitude[i, j, :] = np.sort(
                    ds_nb.vertices_latitude.values[i, j, :]
                )
                ds_nb.vertices_longitude[i, j, :] = np.sort(
                    ds_nb.vertices_longitude.values[i, j, :]
                )

        # Assert all values are close (discard cells at edge of selected grid area)
        xr.testing.assert_allclose(
            ds.lat_vertices.isel({"rlat": range(1, 19), "rlon": range(1, 19)}),
            ds_nb.vertices_latitude.isel({"rlat": range(1, 19), "rlon": range(1, 19)}),
            rtol=1e-06,
            atol=0,
        )
        xr.testing.assert_allclose(
            ds.lon_vertices.isel({"rlat": range(1, 19), "rlon": range(1, 19)}),
            ds_nb.vertices_longitude.isel({"rlat": range(1, 19), "rlon": range(1, 19)}),
            rtol=1e-06,
            atol=0,
        )


def test_calculate_bounds_rectilinear(mini_esgf_data):
    """Test for bounds calculation for rectilinear grid."""

    # Load CORDEX dataset (curvilinear grid)
    with xr.open_dataset(mini_esgf_data["CMIP6_TAS_PRECISION_A"]) as ds:

        # Drop bounds variables and compute them
        ds_nb = ds.drop_vars(["lon_bnds", "lat_bnds"])
        ds_nb = clidu.generate_bounds_rectilinear(ds_nb, lat="lat", lon="lon")

        # Assert all values are close
        xr.testing.assert_allclose(ds.lat_bnds, ds_nb.lat_bnds, rtol=1e-06, atol=0)
        xr.testing.assert_allclose(ds.lon_bnds, ds_nb.lon_bnds, rtol=1e-06, atol=0)


# Adapted from roocs_utils:tests.test_xarray_utils.test_get_main_var
def test_get_main_var(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["C3S_CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        result = clidu.get_main_variable(ds)
        assert result == "tas"


def test_get_main_var_2(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_ZOSTOGA"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        result = clidu.get_main_variable(ds)
        assert result == "zostoga"


def test_get_main_var_3(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        result = clidu.get_main_variable(ds)
        assert result == "tas"


def test_get_main_var_4(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_RH"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        result = clidu.get_main_variable(ds)
        assert result == "rh"


def test_get_main_var_test_data(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP6_SIMASS_DEGEN"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        var_id = clidu.get_main_variable(ds)
        assert var_id == "simass"


def test_get_main_var_include_common_coords(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        var_id = clidu.get_main_variable(ds, exclude_common_coords=False)

        # incorrectly identified main variable and common_coords included in search
        assert var_id == "lat_bnds"


def test_get_standard_names(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        assert sorted(ds.cf.standard_names) == sorted(
            [
                "air_temperature",
                "height",
                "latitude",
                "longitude",
                "time",
            ]
        )


# Adapted from roocs_utils:tests.test_xarray_utils.test_cf_xarray
def test_get_latitude_cf_xarray(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        xr.testing.assert_identical(
            ds["lat"].reset_coords("height", drop=True), ds.cf["lat"]
        )
        xr.testing.assert_identical(
            ds["lat"].reset_coords("height", drop=True), ds.cf["latitude"]
        )


def test_get_latitude_2_cf_xarray(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["C3S_CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        xr.testing.assert_identical(ds["lat"], ds.cf["lat"])
        xr.testing.assert_identical(ds["lat"], ds.cf["latitude"])
        with pytest.raises(KeyError):
            xr.testing.assert_identical(ds["lat"], ds.cf["lats"])


def test_get_lat_lon_names_from_ds_cf_xarray(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        assert ds.cf["latitude"].name == "lat"
        assert ds.cf["longitude"].name == "lon"
        # not sure how it will deal with lats


def test_get_time_cf_xarray(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        xr.testing.assert_identical(
            ds["time"].reset_coords(("height"), drop=True), ds.cf["time"]
        )


# Adapted from roocs_utils:tests.test_xarray_utils.test_get_coords
# test dataset with no known problems
def test_get_time(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        da = ds["tas"]
        coord = da.time
        assert clidu.get_coord_type(coord) == "time"


def test_get_latitude(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        da = ds["tas"]
        coord = da.lat
        assert clidu.get_coord_type(coord) == "latitude"


def test_get_longitude(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        da = ds["tas"]
        coord = da.lon
        assert clidu.get_coord_type(coord) == "longitude"


# test dataset with no standard name for time
def test_get_time_2(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["C3S_CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        da = ds["tas"]
        coord = da.time
        assert clidu.get_coord_type(coord) == "time"


def test_get_latitude_2(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["C3S_CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        da = ds["tas"]
        coord = da.lat
        assert clidu.get_coord_type(coord) == "latitude"


def test_get_longitude_2(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["C3S_CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        da = ds["tas"]
        coord = da.lon
        assert clidu.get_coord_type(coord) == "longitude"


# test dataset with only time and another coordinate that isn't lat or lon
def test_get_time_3(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_ZOSTOGA"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        da = ds["zostoga"]
        coord = da.time
        assert clidu.get_coord_type(coord) == "time"


def test_get_level(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_ZOSTOGA"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        da = ds["zostoga"]
        coord = da.lev
        assert clidu.get_coord_type(coord) == "level"


def test_get_other(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP6_SICONC"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        da = ds["siconc"]
        coord = da.type
        assert clidu.get_coord_type(coord) is None


def test_order_of_coords(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP5_ZOSTOGA"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        da = ds["zostoga"]

        coords = [_ for _ in da.coords]
        assert coords == ["lev", "time"]

        coord_names_keys = [_ for _ in da.coords.keys()]
        assert coord_names_keys == ["lev", "time"]

        # this changes order each time
        # coord_names = [_ for _ in da.coords._names]
        # assert coord_names == ['time', 'lev']

        coord_names_keys = [_ for _ in da.coords]
        assert coord_names_keys == ["lev", "time"]

        coord_sizes = [da[f"{coord}"].size for coord in da.coords.keys()]
        shape = da.shape

        dims = da.dims
        assert dims == ("time", "lev")

        assert shape == (1140, 1)  # looks like shape comes from dims
        assert coord_sizes == [1, 1140]
        assert ds["lev"].shape == (1,)
        assert ds["time"].shape == (1140,)


def test_text_coord_not_level(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["CMIP6_CHAR_DIM"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        coord_type = clidu.get_coord_type(ds.sector)
        assert coord_type is None
        assert coord_type != "level"


def test_get_coords_by_type(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["C3S_CORDEX_AFR_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:

        # check lat, lon, time and level are found when they are coordinates
        lat = clidu.get_coord_by_type(ds, "latitude", ignore_aux_coords=False)
        lon = clidu.get_coord_by_type(ds, "longitude", ignore_aux_coords=False)
        time = clidu.get_coord_by_type(ds, "time", ignore_aux_coords=False)
        level = clidu.get_coord_by_type(ds, "level", ignore_aux_coords=False)

        assert lat == "lat"
        assert lon == "lon"
        assert time == "time"
        assert level == "height"

        # test that latitude and longitude are still found when they are data variables
        # reset coords sets lat and lon as data variables
        ds = ds.reset_coords(["lat", "lon"])

        # if ignore_Aux_coords=True then lat/lon should not be identified
        lat = clidu.get_coord_by_type(ds, "latitude", ignore_aux_coords=True)
        lon = clidu.get_coord_by_type(ds, "longitude", ignore_aux_coords=True)

        assert lat is None
        assert lon is None

        # if ignore_Aux_coords=False then lat/lon should be identified
        lat = clidu.get_coord_by_type(ds, "latitude", ignore_aux_coords=False)
        lon = clidu.get_coord_by_type(ds, "longitude", ignore_aux_coords=False)

        assert lat == "lat"
        assert lon == "lon"


def test_get_coords_by_type_with_no_time(mini_esgf_data):
    with xr.open_mfdataset(
        mini_esgf_data["C3S_CORDEX_AFR_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    ) as ds:
        # check time
        time = clidu.get_coord_by_type(ds, "time", ignore_aux_coords=False)
        assert time == "time"
        # drop time
        ds = ds.drop_dims("time")
        time = clidu.get_coord_by_type(ds, "time", ignore_aux_coords=False)
        assert time is None


# Adapted from roocs_utils:tests.test_xarray_utils.test_open_xr_dataset
def test_open_xr_dataset(mini_esgf_data):
    with clidu.open_xr_dataset(mini_esgf_data["C3S_CMIP5_TAS"]) as ds:
        assert isinstance(ds, xr.Dataset)


@pytest.mark.xfail(
    reason="The xarray issue to yield an empty encoding dictionary for 'time' when calling open_mfdataset seems to have been fixed"
)
def test_open_xr_dataset_retains_time_encoding(mini_esgf_data):
    with clidu.open_xr_dataset(mini_esgf_data["C3S_CMIP5_TAS"]) as ds:
        assert isinstance(ds, xr.Dataset)
        assert hasattr(ds, "time")
        assert ds.time.encoding.get("units") == "days since 1850-01-01 00:00:00"

    # Now test without our clever opener - to prove time encoding is lost
    kwargs = {
        "decoe_times": xr.coders.CFDatetimeCoder(use_cftime=True),
        "decode_timedelta": False,
        "combine": "by_coords",
    }
    with xr.open_mfdataset(glob.glob(mini_esgf_data["C3S_CMIP5_TAS"]), **kwargs) as ds:
        assert ds.time.encoding == {}


def _common_test_open_xr_dataset_kerchunk(uri):
    ds = clidu.open_xr_dataset(uri)
    assert isinstance(ds, xr.Dataset)
    assert "tasmax" in ds

    # Also test time encoding is retained
    assert hasattr(ds, "time")
    assert ds.time.encoding.get("units") == "days since 1850-01-01"

    return ds


@pytest.mark.skip(reason="kerchunk outdated")
def test_open_xr_dataset_kerchunk_json(mini_esgf_data):
    _common_test_open_xr_dataset_kerchunk(
        mini_esgf_data["CMIP6_KERCHUNK_HTTPS_OPEN_JSON"]
    )


@pytest.mark.skip(reason="kerchunk outdated")
def test_open_xr_dataset_kerchunk_zst(mini_esgf_data):
    _common_test_open_xr_dataset_kerchunk(
        mini_esgf_data["CMIP6_KERCHUNK_HTTPS_OPEN_ZST"]
    )


@pytest.mark.skip(reason="kerchunk outdated")
def test_open_xr_dataset_kerchunk_compare_json_vs_zst(mini_esgf_data):
    ds1 = _common_test_open_xr_dataset_kerchunk(
        mini_esgf_data["CMIP6_KERCHUNK_HTTPS_OPEN_JSON"]
    )
    ds2 = _common_test_open_xr_dataset_kerchunk(
        mini_esgf_data["CMIP6_KERCHUNK_HTTPS_OPEN_ZST"]
    )

    diff = ds1.isel(time=slice(0, 2)) - ds2.isel(time=slice(0, 2))
    assert diff.max() == diff.min() == 0.0



import pytest

from clisops.exceptions import InvalidParameterValue
from clisops.parameter import TimeComponentsParameter, string_to_dict, time_components

type_error = (
    "Input type of <{}> not allowed. Must be one of: "
    "[<class 'dict'>, <class 'str'>, <class 'clisops.parameter._utils.TimeComponents'>, <class 'NoneType'>]"
)


tc_str = "year:1999,2000,2001|month:12,01,02|hour:00"
tc_dict = {"year": [1999, 2000, 2001], "month": [12, 1, 2], "hour": [0]}
tc_dict_month_long_names = {
    "year": [1999, 2000, 2001],
    "month": ["December", "January", "February"],
    "hour": [0],
}
tc_dict_short_names = {
    "year": [1999, 2000, 2001],
    "month": ["dec", "jan", "feb"],
    "hour": [0],
}


def test_string_input():
    # year, month, hour
    parameter = TimeComponentsParameter("year:1999,2000,2001|month:dec,jan,feb|hour:00")
    assert parameter.value["year"] == [1999, 2000, 2001]
    assert parameter.value["month"] == [12, 1, 2]
    assert parameter.value["hour"] == [0]
    assert parameter.get_bounds() == ("1999-01-01T00:00:00", "2001-12-31T23:59:59")
    # month
    parameter = TimeComponentsParameter("month:12,1,02")
    assert parameter.value["month"] == [12, 1, 2]
    assert parameter.get_bounds() == (None, None)
    # single year
    parameter = TimeComponentsParameter("year:2010|month:mar,apr,may")
    assert parameter.value["year"] == [2010]
    assert parameter.value["month"] == [3, 4, 5]
    assert parameter.get_bounds() == ("2010-01-01T00:00:00", "2010-12-31T23:59:59")


def test_TimeComponents_class():
    tc1 = time_components(**string_to_dict(tc_str))
    tc2 = time_components(**tc_dict)
    tc3 = time_components(**tc_dict_month_long_names)
    tc4 = time_components(**tc_dict_short_names)

    assert tc1.value == tc2.value
    assert tc2.value == tc3.value
    assert tc2.value == tc4.value


def test__str__():
    parameter = TimeComponentsParameter(tc_str)
    assert str(parameter).startswith("Time components to select:")
    assert "month => [12, 1, 2]" in str(parameter)


def test_raw():
    parameter = TimeComponentsParameter(tc_str)
    assert parameter.raw == tc_str


def test_validate_error_id():
    with pytest.raises(InvalidParameterValue) as exc:
        TimeComponentsParameter("I am rubbish")
    assert str(exc.value) == "Cannot create TimeComponentsParameter from: I am rubbish"


def test_bad_type_input():
    with pytest.raises(InvalidParameterValue) as exc:
        TimeComponentsParameter(34)
    assert str(exc.value) == type_error.format("class 'int'")


def test_dict():
    for input_dct in (tc_dict, tc_dict_short_names, tc_dict_month_long_names):
        parameter = TimeComponentsParameter(input_dct)
        assert parameter.value == tc_dict


def test_time_components_input():
    tc = time_components(**tc_dict)
    parameter = TimeComponentsParameter(tc)
    assert parameter.value == tc_dict


def test_time_components_with_args():
    tc = time_components(year=[200, 500], hour="06")
    assert tc.value["year"] == [200, 500]
    assert tc.value["hour"] == [6]


def test_whitespace():
    parameter = TimeComponentsParameter(tc_str + "   ")
    assert parameter.value == tc_dict


def test_empty_string():
    parameter = TimeComponentsParameter("")
    assert parameter.value is None


def test_none():
    parameter = TimeComponentsParameter(None)
    assert parameter.value is None


def test_class_instance():
    parameter = TimeComponentsParameter(tc_str)
    new_parameter = TimeComponentsParameter(parameter)
    assert new_parameter.value == tc_dict



import os

import pytest

from clisops.utils.file_utils import FileMapper, is_file_list


@pytest.fixture(scope="module")
def cds_domain():
    return "https://data.mips.climate.copernicus.eu"


@pytest.mark.skipif(os.path.isdir("/badc") is False, reason="data not available")
def test_file_mapper():
    file_paths = [
        "/badc/cmip6/data/CMIP6/CMIP/MIROC/MIROC6/amip/r1i1p1f1/day/tas/gn/latest/"
        "tas_day_MIROC6_amip_r1i1p1f1_gn_19790101-19881231.nc",
        "/badc/cmip6/data/CMIP6/CMIP/MIROC/MIROC6/amip/r1i1p1f1/day/tas/gn/latest/"
        "tas_day_MIROC6_amip_r1i1p1f1_gn_19890101-19981231.nc",
    ]
    fm = FileMapper(file_paths)

    assert (
        fm.dirpath
        == "/badc/cmip6/data/CMIP6/CMIP/MIROC/MIROC6/amip/r1i1p1f1/day/tas/gn/latest"
    )
    assert fm.file_paths == file_paths
    assert fm.file_list == [
        "tas_day_MIROC6_amip_r1i1p1f1_gn_19790101-19881231.nc",
        "tas_day_MIROC6_amip_r1i1p1f1_gn_19890101-19981231.nc",
    ]


def test_file_mapper_different_dirpath():
    file_paths = [
        "/badc/cmip6/data/CMIP6/CMIP/MIROC/MIROC6/amip/r1i1p1f1/day/tas/gn/latest/"
        "tas_day_MIROC6_amip_r1i1p1f1_gn_19790101-19881231.nc",
        "/badc/cmip6/data/CMIP6/ScenarioMIP/MIROC/MIROC6/amip/r1i1p1f1/day/tas/gn/latest/"
        "tas_day_MIROC6_amip_r1i1p1f1_gn_19890101-19981231.nc",
    ]

    with pytest.raises(Exception) as exc:
        FileMapper(file_paths)
    assert (
        str(exc.value)
        == "File inputs are not from the same directory so cannot be resolved."
    )


@pytest.mark.skipif(os.path.isdir("/badc") is False, reason="data not available")
def test_file_mapper_fake_files():
    file_paths = [
        "/badc/cmip6/data/CMIP6/ScenarioMIP/MIROC/MIROC6/amip/r1i1p1f1/day/tas/gn/latest/"
        "tas_day_MIROC6_amip_r1i1p1f1_gn_19790101-19881231.nc",
        "/badc/cmip6/data/CMIP6/ScenarioMIP/MIROC/MIROC6/amip/r1i1p1f1/day/tas/gn/latest/"
        "tas_day_MIROC6_amip_r1i1p1f1_gn_19890101-19981231.nc",
    ]

    with pytest.raises(FileNotFoundError) as exc:
        FileMapper(file_paths)
    assert str(exc.value) == "Some files could not be found."


def test_is_file_list(cds_domain):
    coll = ["/badc/cmip6/fake1.nc", "/badc/cmip6/fake2.nc"]
    assert is_file_list(coll) is True

    coll = [
        f"{cds_domain}/cmip6/fake1.nc",
        f"{cds_domain}/badc/cmip6/fake2.nc",
    ]
    assert is_file_list(coll) is True

    coll = ["fake1.nc", "/badc/cmip6/fake2.nc"]
    assert is_file_list(coll) is False

    with pytest.raises(Exception) as exc:
        coll = "/badc/cmip6/fake1.nc"
        is_file_list(coll)
    assert (
        str(exc.value) == "Expected collection as a list, have received <class 'str'>"
    )



import pytest

from clisops.utils.time_utils import AnyCalendarDateTime, str_to_AnyCalendarDateTime


class TestAnyCalendarDateTime:

    def test_simple(self):
        d = AnyCalendarDateTime(1997, 4, 9, 0, 0, 0)
        assert d.value == "1997-04-09T00:00:00"

    def test_str_to_calendar(self):
        d = str_to_AnyCalendarDateTime("1999-01-31")
        d.sub_day()
        assert d.value.startswith("1999-01-30")

        d.add_day()
        d.add_day()
        assert d.value.startswith("1999-02-01")

        d = str_to_AnyCalendarDateTime("1999-02-31")
        d.sub_day()
        assert d.value.startswith("1999-02-30")

        d = str_to_AnyCalendarDateTime("1999-01-01T12:13:14")
        d.sub_day()
        assert d.value == "1998-12-31T12:13:14"

    def test_irregular_datetimes(self):
        # looks like no year
        d = str_to_AnyCalendarDateTime("01-02")
        # first value is taken as the year
        assert d.value == "1-02-01T00:00:00"

        # month is 00
        with pytest.raises(ValueError) as exc:
            str_to_AnyCalendarDateTime("1999-00-33")
        assert (
            str(exc.value)
            == "Invalid input 0 for month. Expected value between 1 and 12."
        )

        # month is 13
        with pytest.raises(ValueError) as exc:
            str_to_AnyCalendarDateTime("1999-13-22")
        assert (
            str(exc.value)
            == "Invalid input 13 for month. Expected value between 1 and 12."
        )

        # hour is 27
        with pytest.raises(ValueError) as exc:
            str_to_AnyCalendarDateTime("1999-01-22T27:00:00")
        assert (
            str(exc.value)
            == "Invalid input 27 for hour. Expected value between 0 and 23."
        )



import numpy as np
import pytest
import xarray as xr
from numpy import array, nan

nat = np.datetime64("NaT")


def test_xarray_da_mean_skipna_true():
    da = xr.DataArray(array([10.0, 10.0, 10.0, 10.0, 10.0, nan, nan, nan, nan, nan]))
    mean = da.mean(skipna=True)
    assert mean != 2
    assert mean == 10
    # skips nans out completely - does not include them when calculating the average


# By default, only skips missing values for float dtypes;
# other dtypes either do not have a sentinel missing value (int) or skipna=True
# has not been implemented (object, datetime64 or timedelta64)
def test_xarray_da_mean_skipna_none_float():
    da = xr.DataArray(array([10.0, 10.0, 10.0, 10.0, 10.0, nan, nan, nan, nan, nan]))
    mean = da.mean(skipna=None)
    assert mean == 10


def test_xarray_da_mean_skipna_false():
    da = xr.DataArray(array([10.0, 10.0, 10.0, 10.0, 10.0, nan, nan, nan, nan, nan]))
    mean = da.mean(skipna=False)
    assert mean != 1
    # result is nan


def test_xarray_da_mean_skipna_true_int():
    da = xr.DataArray(array([10, 10, 10, 10, 10, None, None, None, None, None]))
    mean = da.mean(skipna=True)
    assert mean != 2
    assert mean == 10


def test_xarray_da_mean_skipna_false_int():
    da = xr.DataArray(array([10, 10, 10, 10, 10, None, None, None, None, None]))
    with pytest.raises(TypeError):
        da.mean(skipna=False)


# would expect this to give the same result  as false, but it gives the same result as true
def test_xarray_da_mean_skipna_none_int():
    da = xr.DataArray(array([10, 10, 10, 10, 10, None, None, None, None, None]))
    mean = da.mean(skipna=None)
    assert mean == 10


# nan is a float so means the whole array becomes a float array if nan is used
# means results would the same as above
def test_xarray_da_mean_skipna_true_int_masked():
    x = np.array([10, 10, 10, 10, 10, -1, -1, -1, -1, -1])
    da = xr.DataArray(np.ma.masked_array(x, mask=[0, 0, 0, 0, 0, 1, 1, 1, 1, 1]))
    mean = da.mean(skipna=True)
    assert mean != 2
    assert mean == 10


def test_xarray_da_mean_skipna_none():
    da = xr.DataArray(array([10.0, 10.0, 10.0, 10.0, 10.0, nan, nan, nan, nan, nan]))
    mean = da.mean(skipna=None)
    assert mean != 2
    assert mean == 10


def test_xarray_da_mean_skipna_int_false_masked():
    x = np.array([10, 10, 10, 10, 10, -1, -1, -1, -1, -1])
    da = xr.DataArray(np.ma.masked_array(x, mask=[0, 0, 0, 0, 0, 1, 1, 1, 1, 1]))
    mean = da.mean(skipna=False)
    assert mean != 10
    # result in nan


def test_xarray_da_mean_skipna_true_datetime():
    date = np.datetime64("2010-04-09")
    da = xr.DataArray(array([date, date, date, date, date, nat, nat, nat, nat, nat]))
    mean = da.mean(skipna=True)
    assert mean == np.datetime64("2010-04-09")
    # skips nans out completely - does not include them when calculating the average


def test_xarray_da_mean_skipna_none_datetime():
    date = np.datetime64("2010-04-09")
    da = xr.DataArray(array([date, date, date, date, date, nat, nat, nat, nat, nat]))
    mean = da.mean(skipna=None)
    assert mean == np.datetime64("2010-04-09")


def test_xarray_da_mean_skipna_false_datetime():
    date = np.datetime64("2010-04-09")
    da = xr.DataArray(array([date, date, date, date, date, nat, nat, nat, nat, nat]))
    mean = da.mean(skipna=False)
    assert mean != np.datetime64("2010-04-09")
    # result is nat


def test_xarray_da_mean_keep_attrs_true(mini_esgf_data):
    ds = xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        combine="by_coords",
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        drop_variables=["time_bnds"],
    )
    ds_tas_mean = ds.tas.mean(dim="lat", keep_attrs=True)
    ds_mean = ds.mean(dim="lat", keep_attrs=True)

    assert ds.tas.attrs == ds_tas_mean.attrs
    assert ds.attrs == ds_mean.attrs


def test_xarray_da_mean_keep_attrs_false(mini_esgf_data):
    ds = xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        combine="by_coords",
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
    ).load()
    ds_tas_mean = ds.tas.mean(dim="time", keep_attrs=False)
    ds_mean = ds.mean(dim="time", keep_attrs=False)

    assert ds_tas_mean.attrs == {}
    assert ds_mean.attrs == {}
